# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Extinction functions."""

import numpy as np

__all__ = ['extinction_ccm']

# Optical/NIR coefficients from Cardelli (1989)
c1_ccm = [1., 0.17699, -0.50447, -0.02427, 0.72085, 0.01979, -0.77530, 0.32999]
c2_ccm = [0., 1.41338, 2.28305, 1.07233, -5.38434, -0.62251, 5.30260, -2.09002]
      
# Optical/NIR coefficents from O'Donnell (1994)
c1_odonnell = [1., 0.104, -0.609, 0.701, 1.137, -1.718, -0.827, 1.647, -0.505]
c2_odonnell = [0., 1.952, 2.908, -3.989, -7.985, 11.102, 5.491, -10.805, 3.347]

def extinction_ccm(wavelength, a_v=None, ebv=None, r_v=3.1,
                   optical_coeffs='odonnell'):
    r"""Return the Cardelli, Clayton, and Mathis (1989) extinction curve.

    This function returns the total extinction A(\lambda) at the given
    wavelengths, given either the total V band extinction `a_v` or the
    selective extinction E(B-V) `ebv`, where `a_v = r_v * ebv`.

    Parameters
    ----------
    wavelength : float or list_like
        Wavelength(s) in Angstroms. Values must be between 909.1 and 33,333.3,
        the range of validity of the extinction curve paramterization.
    a_v or ebv: float
        Total V band extinction or selective extinction E(B-V) (must specify
        exactly one).
    r_v : float, optional
        Ratio of total to selective extinction: R_V = A_V / E(B-V).
        Default is 3.1.
    optical_coeffs : {'odonnell', 'ccm'}, optional
        If 'odonnell' (default), use the updated parameters for the optical
        given by O'Donnell (1994) [2]_. If 'ccm', use the original paramters
        given by Cardelli, Clayton and Mathis (1989) [1]_.

    Returns
    -------
    extinction_ratio : float or `~numpy.ndarray`
        Ratio of total to selective extinction: A(wavelengths) / E(B - V)
        at given wavelength(s). 

    Notes
    -----
    In [1]_ the mean :math:`R_V`-dependent extinction law, is parameterized
    as

    .. math::

       <A(\lambda)/A_V> = a(x) + b(x) / R_V

    where the coefficients a(x) and b(x) are functions of
    wavelength. At a wavelength of approximately 5494.5 Angstroms (a
    characteristic wavelength for the V band), a(x) = 1 and b(x) = 0,
    so that A(5494.5 Angstroms) = A_V. This function returns

    .. math::

       A(\lambda) = A_V * (a(x) + b(x) / R_V)

    where `A_V` can either be specified directly or via `E(B-V)`
    (by defintion, `A_V = R_V * E(B-V)`). The flux transmission fraction
    as a function of wavelength can then be obtained by

    .. math::

       T(\lambda) = (10^{-0.4 A(\lambda)})

    The extinction scales linearly with `a_v` or `ebv`, so one can compute
    t ahead of time for a given set of wavelengths with `a_v=1` and then
    scale by `a_v` later:
    `t_base = 10 ** (-0.4 * extinction_ccm(wavelengths, a_v=1.))`, then later:
    `t = np.power(t_base, a_v)`. Similarly for `ebv`.

    For an alternative to the CCM curve, see the extinction curve
    given in Fitzpatrick (1999) [6]_.

    **Notes from the IDL routine CCM_UNRED:**

    1. The CCM curve shows good agreement with the Savage & Mathis (1979)
       [5]_ ultraviolet curve shortward of 1400 A, but is probably
       preferable between 1200 and 1400 A.
    2. Many sightlines with peculiar ultraviolet interstellar extinction 
       can be represented with a CCM curve, if the proper value of 
       R(V) is supplied.
    3. Curve is extrapolated between 912 and 1000 A as suggested by
       Longo et al. (1989) [3]_.
    4. Valencic et al. (2004) [4]_ revise the ultraviolet CCM
       curve (3.3 -- 8.0 um-1).  But since their revised curve does
       not connect smoothly with longer and shorter wavelengths, it is
       not included here.


    References
    ----------
    .. [1] Cardelli, Clayton, and Mathis 1989, ApJ, 345, 245
    .. [2] O'Donnell 1994, ApJ, 422, 158
    .. [3] Longo et al. 1989, ApJ, 339,474
    .. [4] Valencic et al. 2004, ApJ, 616, 912
    .. [5] Savage & Mathis 1979, ARA&A, 17, 73
    .. [6] Fitzpatrick 1999, PASP, 111, 63
    """


    if (a_v is None) and (ebv is None):
        raise ValueError('Must specify either a_v or ebv')
    if (a_v is not None) and (ebv is not None):
        raise ValueError('Cannot specify both a_v and ebv')
    if a_v is None:
        a_v = r_v * ebv

    wavelength = np.asarray(wavelength)
    in_ndim = wavelength.ndim

    x = 1.e4 / wavelength.ravel()  # Inverse microns.
    if ((x < 0.3) | (x > 11.)).any():
        raise ValueError("extinction only defined in wavelength range"
                         " [909.091, 33333.3].")

    a = np.empty(x.shape, dtype=np.float)
    b = np.empty(x.shape, dtype=np.float)
  
    # Infrared
    idx = x < 1.1
    if idx.any():
        a[idx] = 0.574 * x[idx] ** 1.61
        b[idx] = -0.527 * x[idx] ** 1.61

    # Optical/NIR
    idx = (x >= 1.1) & (x < 3.3)
    if idx.any():
        xp = x[idx] - 1.82

        if optical_coeffs == 'odonnell':
            c1, c2 = c1_odonnell, c2_odonnell
        elif optical_coeffs == 'ccm':
            c1, c2 = c1_ccm, c2_ccm
        else:
            raise ValueError('Unrecognized optical_coeffs: {0!r}'
                             .format(optical_coeffs))

        # we need to flip the coefficients, because in polyval
        # c[0] corresponds to x^(N-1), but above, c[0] corresponds to x^0
        a[idx] = np.polyval(np.flipud(c1), xp)
        b[idx] = np.polyval(np.flipud(c2), xp)

    # Mid-UV
    idx = (x >= 3.3) & (x < 8.)
    if idx.any():
        xp = x[idx]
        a[idx] = 1.752 - 0.316 * xp - 0.104 / ((xp - 4.67)**2 + 0.341)
        b[idx] = -3.090 + 1.825 * xp + 1.206 / ((xp - 4.67)**2 + 0.263)

    idx = (x > 5.9) & (x < 8.)
    if idx.any():
        xp = x[idx] - 5.9
        a[idx] += -0.04473 * xp**2 - 0.009779 * xp**3
        b[idx] += 0.2130 * xp**2 + 0.1207 * xp**3

    # Far-UV
    idx = x >= 8.
    if idx.any():
        xp = x[idx] - 8.
        c1 = [ -1.073, -0.628,  0.137, -0.070]
        c2 = [ 13.670,  4.257, -0.420,  0.374]
        a[idx] = np.polyval(np.flipud(c1), xp)
        b[idx] = np.polyval(np.flipud(c2), xp)

    extinction = (a + b / r_v) * a_v
    if in_ndim == 0:
        return extinction[0]
    return extinction



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Simple implementation of nested sampling routine to evaluate Bayesian
evidence."""

import math
import time
from sys import stdout

import numpy as np

def randsphere(n):
    """Draw a random point within a n-dimensional unit sphere"""

    z = np.random.randn(n)
    return z * np.random.rand()**(1./n) / np.sqrt(np.sum(z**2))

def ellipsoid(X, expand=1.):
    """
    Calculate ellipsoid containing all samples X.

    Parameters
    ----------
    X : (nobj, ndim) ndarray
        Coordinates of points.
    expand : float, optional.
        Expand the ellipsoid by this linear factor. Default is 1, which
        corresponds to an ellipsoid that just barely encloses all the points.
        Note that volume is increased by a factor of ``(expand)**ndim``

    Returns
    -------
    vs : (ndim, ndim) ndarray
        Scaled eigenvectors (in columns): vs[:,i] is the i-th eigenvector.
    mean : (ndim,) ndarray
        Simple average of all samples.

    Notes
    -----
    For the 2-d case, to verify that the generated ellipse encloses all
    the points, the ellipse can be plotted using matplotlib on an existing
    Axes  ``ax`` as follows:

        from matplotlib.patches import Ellipse

        width = np.sqrt(np.sum(vs[:,1]**2)) * 2.
        height = np.sqrt(np.sum(vs[:,0]**2)) * 2.
        angle = math.atan(vs[1,1] / vs[0,1]) * 180./math.pi
        e = Ellipse(mean, width, height, angle)
        e.set_facecolor('None')
        ax.add_artist(e)

    To draw the vectors ``vs``:
    
        for i in [0,1]:
            plt.arrow(mean[0], mean[1], vs[0, i], vs[1, i])
    """

    X_avg = np.mean(X, axis=0)
    Xp = X - X_avg
    c = np.cov(Xp, rowvar=0)
    cinv = np.linalg.inv(c)
    w, v = np.linalg.eig(c)
    vs = np.dot(v, np.diag(np.sqrt(w)))  # scaled eigenvectors

    # Calculate 'k' factor
    k = np.empty(len(X), dtype=np.float)

    #for i in range(len(k)):
    #    k[i] = np.dot(np.dot(Xp[i,:], cinv), Xp[i,:])
    
    # equivalent to above:
    tmp = np.tensordot(Xp, cinv, axes=1)
    for i in range(len(k)):
        k[i] = np.dot(tmp[i,:], Xp[i,:])

    k = np.max(k)

    return np.sqrt(k) * expand * vs, X_avg

def sample_ellipsoid(vs, mean, nsamples=1):
    """Chose sample(s) randomly distributed within an ellipsoid.
    
    Parameters
    ----------
    vs : (ndim, ndim) ndarray
        Scaled eigenvectors (in columns): vs[:,i] is the i-th eigenvector.
    mean : (ndim,) ndarray
        Simple average of all samples.

    Returns
    -------
    x : (nsamples, ndim) array, or (ndim,) array when nsamples == 1
        Coordinates within the ellipsoid.
    """

    ndim = len(mean)
    if nsamples == 1:
        return np.dot(vs, randsphere(ndim)) + mean

    x = np.empty((nsamples, ndim), dtype=np.float)
    for i in range(nsamples):
        x[i, :] = np.dot(vs, randsphere(ndim)) + mean
    return x

def nest(loglikelihood, prior, npar, nobj=50, maxiter=10000,
         verbose=False, verbose_name=''):
    """Simple nested sampling algorithm to evaluate Bayesian evidence.

    Parameters
    ----------
    loglikelihood : func
        Function returning log(likelihood) given parameters as a 1-d numpy
        array of length `npar`. 
    prior : func
        Function translating a unit cube to the parameter space according to 
        the prior. The input is a 1-d numpy array with length `npar`, where
        each value is in the range [0, 1). The return value should also be a
        1-d numpy array with length `npar`, where each value is a parameter.
        The return value is passed to the loglikelihood function. For example,
        for a 2 parameter model with flat priors in the range [0, 2), the
        function would be

            def prior(u):
                return 2. * u

    npar : int
        Number of parameters.
    nobj : int, optional
        Number of random samples. Larger numbers result in a more finely
        sampled posterior (more accurate evidence), but also a larger
        number of iterations required to converge. Default is 50.
    maxiter : int, optional
        Maximum number of iterations. Iteration may stop earlier if
        termination condition is reached. Default is 10000. The total number
        of likelihood evaluations will be ``nexplore * niter``.
    verbose : bool, optional
        Print a single line of running total iterations.
    verbose_name : str, optional
        Print this string at start of the iteration line printed when
        verbose=True.

    Returns
    -------
    results : dict
        Containing following keys:

        * `niter` (int) number of iterations.
        * `ncalls` (int) number of likelihood calls.
        * `time` (float) time in seconds.
        * `logz` (float) log of evidence.
        * `logzerr` (float) error on `logz`.
        * `loglmax` (float) Maximum likelihood of any sample.
        * `h` (float) information.
        * `samples_parvals` (array, shape=(nsamples, npar)) parameter values
          of each sample.
        * `samples_wt` (array, shape=(nsamples,) Weight of each sample.

    Notes
    -----
    This is an implementation of John Skilling's Nested Sampling algorithm,
    following the ellipsoidal sampling algorithm in Shaw et al (2007). Only a
    single ellipsoid is used.
    
    Sample Weights are ``likelihood * prior_vol`` where
    prior_vol is the fraction of the prior volume the sample represents.

    References
    ----------
    http://www.inference.phy.cam.ac.uk/bayesys/
    Shaw, Bridges, Hobson 2007, MNRAS, 378, 1365
    """

    # Initialize objects and calculate likelihoods
    objects_u = np.random.random((nobj, npar)) #position in unit cube
    objects_v = np.empty((nobj, npar), dtype=np.float) #position in unit cube
    objects_logl = np.empty(nobj, dtype=np.float)  # log likelihood
    for i in range(nobj):
        objects_v[i,:] = prior(objects_u[i,:])
        objects_logl[i] = loglikelihood(objects_v[i,:])

    # Initialize values for nested sampling loop.
    samples_parvals = [] # stored objects for posterior results
    samples_logwt = []
    loglstar = None  # ln(Likelihood constraint)
    h = 0.  # Information, initially 0.
    logz = -1.e300  # ln(Evidence Z, initially 0)
    # ln(width in prior mass), outermost width is 1 - e^(-1/n)
    logwidth = math.log(1. - math.exp(-1./nobj))
    loglcalls = nobj #number of calls we already made

    # Nested sampling loop.
    ndecl = 0
    logwt_old = None
    time0 = time.time()
    for it in range(maxiter):
        if verbose:
            if logz > -1.e6:
                print "\r{} iter={:6d} logz={:8f}".format(verbose_name, it,
                                                          logz),
            else:
                print "\r{} iter={:6d} logz=".format(verbose_name, it),
            stdout.flush()

        # worst object in collection and its weight (= width * likelihood)
        worst = np.argmin(objects_logl)
        logwt = logwidth + objects_logl[worst]

        # update evidence Z and information h.
        logz_new = np.logaddexp(logz, logwt)
        h = (math.exp(logwt - logz_new) * objects_logl[worst] +
             math.exp(logz - logz_new) * (h + logz) -
             logz_new)
        logz = logz_new

        # Add worst object to samples.
        samples_parvals.append(np.array(objects_v[worst]))
        samples_logwt.append(logwt)

        # The new likelihood constraint is that of the worst object.
        loglstar = objects_logl[worst]

        # calculate the ellipsoid in parameter space that contains all the
        # samples (including the worst one).
        vs, mean = ellipsoid(objects_u, expand=1.06)

        # choose a point from within the ellipse until it has likelihood
        # better than loglstar
        while True:
            u = sample_ellipsoid(vs, mean)
            if np.any(u < 0.) or np.any(u > 1.):
                continue
            v = prior(u)
            logl = loglikelihood(v)
            loglcalls += 1

            # Accept if and only if within likelihood constraint.
            if logl > loglstar:
                objects_u[worst] = u
                objects_v[worst] = v
                objects_logl[worst] = logl
                break

        # Shrink interval
        logwidth -= 1./nobj

        # stop when the logwt has been declining for more than 10 or niter/4
        # consecutive iterations.
        if logwt < logwt_old:
            ndecl += 1
        else:
            ndecl = 0
        if ndecl > 10 and ndecl > it / 6:
            break
        logwt_old = logwt

    tottime = time.time() - time0
    if verbose:
        print 'calls={:d} time={:7.3f}s'.format(loglcalls, tottime)

    # Add remaining objects.
    # After N samples have been taken out, the remaining width is e^(-N/nobj)
    # The remaining width for each object is e^(-N/nobj) / nobj
    # The log of this for each object is:
    # log(e^(-N/nobj) / nobj) = -N/nobj - log(nobj)
    logwidth = -len(samples_parvals) / nobj - math.log(nobj)
    for i in range(nobj):
        logwt = logwidth + objects_logl[i]
        logz_new = np.logaddexp(logz, logwt)
        h = (math.exp(logwt - logz_new) * objects_logl[i] +
             math.exp(logz - logz_new) * (h + logz) -
             logz_new)
        logz = logz_new
        samples_parvals.append(np.array(objects_v[i]))
        samples_logwt.append(logwt)

    return {
        'niter': it + 1,
        'ncalls': loglcalls,
        'time': tottime,
        'logz': logz,
        'logzerr': math.sqrt(h / nobj),
        'loglmax': np.max(objects_logl),
        'h': h,
        'samples_parvals': np.array(samples_parvals),  #(nsamp, npar)
        'samples_wt':  np.exp(np.array(samples_logwt) - logz)  #(nsamp,)
        }



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to plot light curve data and models."""
from __future__ import division

import math
import numpy as np

from .spectral import get_bandpass, get_magsystem
from .fitting import normalized_flux

__all__ = ['plotlc']

def normalized_flux(data, zp=25., magsys='ab'):
    """Return flux values normalized to a common zeropoint and magnitude
    system."""

    datalen = len(data['flux'])
    magsys = get_magsystem(magsys)
    flux = np.empty(datalen, dtype=np.float)
    fluxerr = np.empty(datalen, dtype=np.float)

    for i in range(datalen):
        ms = get_magsystem(data['zpsys'][i])
        factor = (ms.zpbandflux(data['band'][i]) /
                  magsys.zpbandflux(data['band'][i]) *
                  10.**(0.4 * (zp - data['zp'][i])))
        flux[i] = data['flux'][i] * factor
        fluxerr[i] = data['fluxerr'][i] * factor

    return flux, fluxerr
                       
def plotlc(data, fname=None, model=None, show_pulls=True,
           include_model_error=False, xfigsize=None, yfigsize=None, dpi=100):
    """Plot light curve data.

    Parameters
    ----------
    data : `~numpy.ndarray` or dict thereof
        Structured array or dictionary of arrays, with the following fields:
        {'time', 'band', 'flux', 'fluxerr', 'zp', 'zpsys'}.
    fname : str, optional
        Filename to write plot to. If `None` (default), plot is shown using
        ``show()``.
    model : `~sncosmo.Model`, optional
        If given, model light curve is overplotted.
    show_pulls : bool, optional
        If True (and if model is given), plot pulls. Default is ``True``.
    include_model_error : bool, optional
        Plot model error as a band around the model.
    xfigsize, yfigsize : float, optional
        figure size in inches in x or y. Specify one or the other, not both.
        Default is xfigsize=8.
    dpi : float, optional
        dpi to pass to ``plt.savefig()`` for rasterized images. 
        
    Examples
    --------

    Suppose we have data in a file that looks like this::
 
        time band flux fluxerr zp zpsys
        55070.0 sdssg -0.263064256628 0.651728140824 25.0 ab
        55072.0512821 sdssr -0.836688186816 0.651728140824 25.0 ab
        55074.1025641 sdssi -0.0104080573938 0.651728140824 25.0 ab
        55076.1538462 sdssz -0.0794771107707 0.651728140824 25.0 ab
        55078.2051282 sdssg 0.897840283912 0.651728140824 25.0 ab
        ...

    To read and plot the data:

        >>> meta, data = sncosmo.readlc('mydatafile.dat')  # read the data
        >>> sncosmo.plotlc(data, fname='plotlc_example.png')  # plot the data

    We happen to know the model and parameters that fit this
    data. Specifying the ``model`` keyword will plot the model over
    the data.
    
        >>> model = sncosmo.get_model('salt2')
        >>> model.set(z=0.5, c=0.2, t0=55100., mabs=-19.5, x1=0.5)
        >>> sncosmo.plotlc(data, fname='plotlc_example.png', model=model)

    .. image:: /pyplots/plotlc_example.png

    """

    import matplotlib.pyplot as plt
    from matplotlib import cm
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    cmap = cm.get_cmap('gist_rainbow')
    disprange = (3000., 10000.)

    dataflux, datafluxerr = normalized_flux(data, zp=25., magsys='ab')

    bandnames = np.unique(data['band']).tolist()
    bands = [get_bandpass(bandname) for bandname in bandnames]
    disps = [b.disp_eff for b in bands]

    # Calculate layout of figure (columns, rows, figure size)
    nsubplots = len(bands)
    ncol = 2
    nrow = (nsubplots - 1) // ncol + 1
    if xfigsize is None and yfigsize is None:
        figsize = (4. * ncol, 3. * nrow)
    elif yfigsize is None:
        figsize = (xfigsize, 3. / 4. * nrow / ncol * xfigsize)
    elif xfigsize is None:
        figsize = (4. / 3. * ncol / nrow * yfigsize, yfigsize)
    else:
        raise ValueError('cannot specify both xfigsize and yfigsize')
    fig = plt.figure(figsize=figsize)

    axnum = 0
    for disp, band, bandname in sorted(zip(disps, bands, bandnames)):
        axnum += 1

        idx = data['band'] == bandname
        time = data['time'][idx]
        flux = dataflux[idx]
        fluxerr = datafluxerr[idx]

        color = cmap((disprange[1] - disp) / (disprange[1] - disprange[0]))

        ax = plt.subplot(nrow, ncol, axnum)
        plt.text(0.9, 0.9, bandname, color='k', ha='right', va='top',
                 transform=ax.transAxes)
        if axnum % 2:
            plt.ylabel('flux ($ZP_{AB} = 25$)')

        if model is None:
            plt.errorbar(time, flux, fluxerr, ls='None',
                         color=color, marker='.', markersize=3.)

        if model is not None and model.bandoverlap(band):
            t0 = model.params['t0']
            plt.errorbar(time - t0, flux, fluxerr, ls='None',
                         color=color, marker='.', markersize=3.)

            result = model.bandflux(band, zp=25., zpsys='ab',
                                    include_error=include_model_error)
            if include_model_error:
                modelflux, modelfluxerr = result
            else:
                modelflux = result

            plt.plot(model.times() - t0, modelflux, ls='-', marker='None',
                     color=color)
            if include_model_error:
                plt.fill_between(model.times() - t0, modelflux - modelfluxerr,
                                 modelflux + modelfluxerr, color=color,
                                 alpha=0.2)


            # steal part of the axes and plot pulls
            if show_pulls:
                divider = make_axes_locatable(ax)
                axpulls = divider.append_axes("bottom", size=0.7, pad=0.1,
                                              sharex=ax)
                modelflux = model.bandflux(band, time, zp=25., zpsys='ab') 
                pulls = (flux - modelflux) / fluxerr
                plt.plot(time - t0, pulls, marker='.', markersize=5.,
                         color=color, ls='None')
                plt.axhline(y=0., color=color)
                plt.setp(ax.get_xticklabels(), visible=False)
                plt.xlabel('time - {:.2f}'.format(t0))
                if axnum % 2:
                    plt.ylabel('pull')

            # maximum plot range
            ymin, ymax = ax.get_ylim()
            maxmodelflux = modelflux.max()
            ymin = max(ymin, -0.2 * maxmodelflux)
            ymax = min(ymax, 2. * maxmodelflux)
            ax.set_ylim(ymin, ymax)

    plt.subplots_adjust(left=0.1, right=0.95, bottom=0.1, top=0.97,
                        wspace=0.2, hspace=0.2)
    if fname is None:
        plt.show()
    else:
        plt.savefig(fname, dpi=dpi)
        plt.clf()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Tests for extinction curve."""

import numpy as np
from sncosmo.extinction import extinction_ccm

def test_extinction_ccm_shapes():

    # Test single value
    extinction_ccm(1.e4, a_v=1.)

    # multiple values
    assert extinction_ccm([1.e4], a_v=1.).shape == (1,)
    assert extinction_ccm([1.e4, 2.e4], a_v=1.).shape == (2,)


def test_extinction_ccm_values():

    r_v = 3.1

    # U, B, V, R, I, J, H, K band effective wavelengths from CCM '89 table 3
    x_inv_microns = np.array([2.78, 2.27, 1.82, 1.43, 1.11, 0.80, 0.63, 0.46])

    # A(lambda)/A(V) for R_V = 3.1 from Table 3 of CCM '89
    ccm_ratio_true = np.array([1.569, 1.337, 1.000, 0.751, 0.479, 0.282,
                               0.190, 0.114])

    wavelengths = 1.e4 / x_inv_microns  # wavelengths in Angstroms
    ccm_ratio = extinction_ccm(wavelengths, r_v=3.1, a_v=1.,
                               optical_coeffs='ccm')

    # TODO:
    # So far, these are close but not exact.
    # I get: [ 1.56880904  1.32257836  1. 0.75125994  0.4780346   0.28206957
    #          0.19200814  0.11572348]

    # At the sigfigs of Table 3, the differences are:
    # [ None, 0.014, None, None, 0.001, None, 0.002, 0.002 ]
    # with B band being the most significant difference.

    # Could be a floating point issue, due to the polynomials? Maybe in the 
    # original Cardelli paper? Should compare a and b to the values in the
    # table.



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .spectral import *
from .models import *
from .lcio import *
from .plotting import *
from .fitting import *
from .typing import *
from .extinction import *
from . import registry
from . import _builtin

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''
try:
    from .version import githash as __githash__
except ImportError:
    __githash__ = ''

# set up the test command
def _get_test_runner():
    from astropy.tests.helper import TestRunner
    return TestRunner(__path__[0])

def test(package=None, test_path=None, args=None, plugins=None,
         verbose=False, pastebin=None, remote_data=False, pep8=False,
         pdb=False, coverage=False, **kwargs):
    """
    Run the tests using py.test. A proper set of arguments is constructed and
    passed to `pytest.main`.

    Parameters
    ----------
    package : str, optional
        The name of a specific package to test, e.g. 'io.fits' or 'utils'.
        If nothing is specified all default tests are run.

    test_path : str, optional
        Specify location to test by path. May be a single file or
        directory. Must be specified absolutely or relative to the
        calling directory.

    args : str, optional
        Additional arguments to be passed to `pytest.main` in the `args`
        keyword argument.

    plugins : list, optional
        Plugins to be passed to `pytest.main` in the `plugins` keyword
        argument.

    verbose : bool, optional
        Convenience option to turn on verbose output from py.test. Passing
        True is the same as specifying `-v` in `args`.

    pastebin : {'failed','all',None}, optional
        Convenience option for turning on py.test pastebin output. Set to
        'failed' to upload info for failed tests, or 'all' to upload info
        for all tests.

    remote_data : bool, optional
        Controls whether to run tests marked with @remote_data. These
        tests use online data and are not run by default. Set to True to
        run these tests.

    pep8 : bool, optional
        Turn on PEP8 checking via the pytest-pep8 plugin and disable normal
        tests. Same as specifying `--pep8 -k pep8` in `args`.

    pdb : bool, optional
        Turn on PDB post-mortem analysis for failing tests. Same as
        specifying `--pdb` in `args`.

    coverage : bool, optional
        Generate a test coverage report.  The result will be placed in
        the directory htmlcov.

    kwargs
        Any additional keywords passed into this function will be passed
        on to the astropy test runner.  This allows use of test-related
        functionality implemented in later versions of astropy without
        explicitly updating the package template.

    See Also
    --------
    pytest.main : py.test function wrapped by `run_tests`.

    """
    test_runner = _get_test_runner()
    return test_runner.run_tests(
        package=package, test_path=test_path, args=args,
        plugins=plugins, verbose=verbose, pastebin=pastebin,
        remote_data=remote_data, pep8=pep8, pdb=pdb,
        coverage=coverage, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Importing this module registers loaders for built-in magnitude systems.
# The module docstring, a table of the magnitude systems, is generated at the
# after all the magnitude systems are registered.

import string

from astropy.io import fits
from astropy.utils.data import download_file
from astropy import units as u

from .. import registry
from .. import Spectrum, MagSystem, SpectralMagSystem, ABMagSystem

# ---------------------------------------------------------------------------
# AB system

def load_ab(name=None):
    return ABMagSystem(name=name)

registry.register_loader(
    MagSystem, 'ab', load_ab, subclass='`~sncosmo.ABMagSystem`',
    description='Source of 3631 Jy has magnitude 0 in all bands')


# ---------------------------------------------------------------------------
# Spectral systems

def load_spectral_magsys_fits(remote_url, name=None):
    fn = download_file(remote_url, cache=True)
    hdulist = fits.open(fn)
    dispersion = hdulist[1].data['WAVELENGTH']
    flux_density = hdulist[1].data['FLUX']
    hdulist.close()
    refspectrum = Spectrum(dispersion, flux_density, 
                           unit=(u.erg / u.s / u.cm**2 / u.AA), dunit=u.AA)
    return SpectralMagSystem(refspectrum, name=name)

calspec_url = 'ftp://ftp.stsci.edu/cdbs/current_calspec/'
vega_url = calspec_url + 'alpha_lyr_stis_005.fits'
registry.register_loader(
    MagSystem, 'vega', load_spectral_magsys_fits, [vega_url],
    subclass='`~sncosmo.SpectralMagSystem`', url=calspec_url,
    description='Vega (alpha lyrae) has magnitude 0 in all bands')

bd17_url = calspec_url + 'bd_17d4708_stisnic_003.fits'
registry.register_loader(
    MagSystem, 'bd17', load_spectral_magsys_fits, [bd17_url],
    subclass='`~sncosmo.SpectralMagSystem`', url=calspec_url,
    description='BD+17d4708 has magnitude 0 in all bands.')

# --------------------------------------------------------------------------
# Generate docstring

lines = ['',
         '  '.join([10*'=', 60*'=', 35*'=', 15*'=']),
         '{:10}  {:60}  {:35}  {:15}'
         .format('Name', 'Description', 'Subclass', 'Spectrum Source')]
lines.append(lines[1])

urlnums = {}
for m in registry.get_loaders_metadata(MagSystem):

    urllink = ''
    description = ''

    if 'description' in m:
        description = m['description']

    if 'url' in m:
        url = m['url']
        if url not in urlnums:
            if len(urlnums) == 0: urlnums[url] = 0
            else: urlnums[url] = max(urlnums.values()) + 1
        urllink = '`{}`_'.format(string.letters[urlnums[url]])

    lines.append("{0!r:10}  {1:60}  {2:35}  {3:15}"
                 .format(m['name'], description, m['subclass'], urllink))

lines.extend([lines[1], ''])
for url, urlnum in urlnums.iteritems():
    lines.append('.. _`{}`: {}'.format(string.letters[urlnum], url))
lines.append('')
__doc__ = '\n'.join(lines)

del lines
del urlnums



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

def _stripcomment(line, char='#'):
    pos = line.find(char)
    if pos == -1: return line
    else: return line[:pos]

def read_griddata(name_or_obj):
    """Read 2-d grid data from a text file.

    Each line has values `x0 x1 y`. Space separated.
    x1 values are only read for first x0 value. Others are assumed
    to match.

    Parameters
    ----------
    filename : str or file-like object

    Returns
    -------
    x0 : numpy.ndarray
        1-d array.
    x1 : numpy.ndarray
        1-d array.
    y : numpy.ndarray
        2-d array of shape (len(x0), len(x1)).
    """

    if isinstance(name_or_obj, basestring):
        f = open(filename, 'rb')
    else:
        f = name_or_obj

    x0 = []    # x0 values.
    x1 = None  # x1 values for first x0 value, assume others are the same.
    y = []     # 2-d array of internal values

    x0_current = None
    x1_current = []
    y1_current = []
    for line in f:
        stripped_line = _stripcomment(line)
        if len(stripped_line) == 0: continue
        x0_tmp, x1_tmp, y_tmp = map(float, stripped_line.split())
        if x0_current is None: x0_current = x0_tmp  #Initialize first time

        # If there is a new x0 value, ingest the old one and reset values
        if x0_tmp != x0_current:
            x0.append(x0_current)
            if x1 is None: x1 = x1_current
            y.append(y1_current)

            x0_current = x0_tmp
            x1_current = []
            y1_current = []

        x1_current.append(x1_tmp)
        y1_current.append(y_tmp)

    # Ingest the last x0 value and y1 array
    x0.append(x0_current)
    y.append(y1_current)

    f.close()
    return np.array(x0), np.array(x1), np.array(y)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import abc
from copy import deepcopy

import numpy as np

from astropy.utils import OrderedDict
from astropy.utils.misc import lazyproperty
import astropy.units as u
from astropy import cosmology
import astropy.constants as const

from . import registry

__all__ = ['get_bandpass', 'get_magsystem', 'Bandpass', 'Spectrum',
           'MagSystem', 'SpectralMagSystem', 'ABMagSystem']

def get_bandpass(name):
    """Get a Bandpass from the registry by name."""
    return registry.retrieve(Bandpass, name)

def get_magsystem(name):
    """Get a MagSystem from the registery by name."""
    return registry.retrieve(MagSystem, name)


class Bandpass(object):
    """Transmission as a function of spectral dispersion.

    Parameters
    ----------
    disp : list_like
        Dispersion. Monotonically increasing values.
    trans : list_like
        Transmission fraction.
    dunit : `~astropy.units.Unit` or str, optional
        Dispersion unit. Default is Angstroms.
    name : str, optional
        Identifier. Default is `None`.

    Examples
    --------
    >>> band = get_bandpass('sdssg')
    >>> band.disp  # wavelengths in Angstroms
    >>> band.ddisp # np.gradient(band.disp)
    >>> band.trans  # transmisson fraction corresponding to wavelengths
    >>> band.name  # 'sdssg'
    """

    def __init__(self, disp, trans, dunit=u.AA, name=None):
        
        disp = np.asarray(disp) 
        trans = np.asarray(trans)
        if disp.shape != trans.shape:
            raise ValueError('shape of disp and trans must match')
        if disp.ndim != 1:
            raise ValueError('only 1-d arrays supported')

        if dunit is not u.AA:
            dunit = u.Unit(dunit)
            disp = dunit.to(u.AA, disp, u.spectral())

        if disp[0] > disp[-1]:
            disp = np.flipud(disp)
            trans = np.flipud(trans)

        self.disp = disp
        self.ddisp = np.gradient(disp)
        self.trans = trans
        self.name = name
        
    @lazyproperty
    def disp_eff(self):
        """Effective wavelength of bandpass in Angstroms."""
        weights = self.trans * np.gradient(self.disp) 
        return np.sum(self.disp * weights) / np.sum(weights)

    def to_unit(self, unit):
        """Return dispersion and transmission in new dispersion units.

        If the requested units are the same as the current units, self is
        returned.

        Parameters
        ----------
        unit : `~astropy.units.Unit` or str
            Target dispersion unit.

        Returns
        -------
        disp : `~numpy.ndarray`
        trans : `~numpy.ndarray`
        """

        if unit is u.AA:
            return self
        
        d = u.AA.to(unit, self.disp, u.spectral())
        t = self.trans
        if d[0] > d[-1]:
            d = np.flipud(d)
            t = np.flipud(t)
        return d, t

    def __repr__(self):
        name = ''
        if self.name is not None:
            name = ' {0!r:s}'.format(self.name)
        return "<Bandpass{0:s} at 0x{1:x}>".format(name, id(self))


class Spectrum(object):
    """A spectrum, representing dispersion and spectral density values.

    Parameters
    ----------
    disp : list_like
        Dispersion values.
    flux : list_like
        Spectral flux density values.
    error : list_like, optional
        1 standard deviation uncertainty on flux density values.
    dunit : `~astropy.units.Unit`
        Units 
    units : `~astropy.units.BaseUnit`
        For now, only units with flux density in energy (not photon counts).
    z : float, optional
        Redshift of spectrum (default is `None`)
    dist : float, optional
        Luminosity distance in Mpc, used to adjust flux upon redshifting.
        The default is ``None``.
    meta : OrderedDict, optional
        Metadata.
    """

    def __init__(self, disp, flux, error=None,
                 unit=(u.erg / u.s / u.cm**2 / u.AA), dunit=u.AA,
                 z=None, dist=None, meta=None):
        
        self._disp = np.asarray(disp)
        self._flux = np.asarray(flux)
        self._dunit = dunit
        self._unit = unit
        self._z = z
        self._dist = dist

        if error is not None:
            self._error = np.asarray(error)
            if self._disp.shape != self._error.shape:
                raise ValueError('shape of wavelength and variance must match')
        else:
            self._error = None

        if meta is None:
            self.meta = OrderedDict()
        else:
            self.meta = deepcopy(meta)

        if self._disp.shape != self._flux.shape:
            raise ValueError('shape of wavelength and flux must match')
        if self._disp.ndim != 1:
            raise ValueError('only 1-d arrays supported')


    @property
    def disp(self):
        """Dispersion values."""
        return self._disp
        
    @property
    def flux(self):
        """Spectral flux density values"""
        return self._flux

    @property
    def error(self):
        """Uncertainty on flux density."""
        return self._error

    @property
    def dunit(self):
        """Units of dispersion."""
        return self._dunit
        
    @property
    def unit(self):
        """Units of flux density."""
        return self._unit

    @property
    def z(self):
        """Redshift of spectrum."""
        return self._z

    @z.setter
    def z(self, value):
        self._z = value

    @property
    def dist(self):
        """Distance to object in Mpc."""
        return self._dist

    @dist.setter
    def dist(self, value):
        self._dist = value


    def bandflux(self, band):
        """Perform synthentic photometry in a given bandpass.
      
        The bandpass transmission is interpolated onto the dispersion grid
        of the spectrum. The result is a weighted sum of the spectral flux
        density values (weighted by transmission values).
        
        Parameters
        ----------
        band : Bandpass object or name of registered bandpass.

        Returns
        -------
        bandflux : float
            Total flux in ph/s/cm^2. If part of bandpass falls
            outside the spectrum, `None` is returned instead.
        bandfluxerr : float
            Error on flux. Only returned if the `error` attribute is not
            `None`.
        """

        band = get_bandpass(band)
        bdisp, btrans = band.to_unit(self._dunit)

        if (bdisp[0] < self._disp[0] or
            bdisp[-1] > self._disp[-1]):
            return None

        idx = ((self._disp > bdisp[0]) & 
               (self._disp < bdisp[-1]))
        d = self._disp[idx]
        f = self._flux[idx]

        #TODO: use spectral density equivalencies once they can do photons.
        # first convert to ergs / s /cm^2 / (dispersion unit)
        target_unit = u.erg / u.s / u.cm**2 / self._dunit
        if self._unit != target_unit:
            f = self._unit.to(target_unit, f, 
                              u.spectral_density(self._dunit, d))

        # Then convert ergs to photons: photons = Energy / (h * nu)
        f = f / const.h.cgs.value / self._dunit.to(u.Hz, d, u.spectral())

        trans = np.interp(d, bdisp, btrans)
        binw = np.gradient(d)
        ftot = np.sum(f * trans * binw)

        if self._error is None:
            return ftot

        else:
            e = self._error[idx]

            # Do the same conversion as above
            if self._unit != target_unit:
                e = self._unit.to(target_unit, e, 
                                  u.spectral_density(self._dunit, d))
            e = e / const.h.cgs.value / self._dunit.to(u.Hz, d, u.spectral())
            etot = np.sqrt(np.sum((e * binw) ** 2 * trans))
            return ftot, etot



    def redshifted_to(self, z, adjust_flux=False, dist=None, cosmo=None):
        """Return a new Spectrum object at a new redshift.

        The current redshift must be defined (self.z cannot be `None`).
        A factor of (1 + z) / (1 + self.z) is applied to the wavelength. 
        The inverse factor is applied to the flux so that the bolometric
        flux (e.g., erg/s/cm^2) remains constant.
        
        .. note:: Currently this only works for units in ergs.

        Parameters
        ----------
        z : float
            Target redshift.
        adjust_flux : bool, optional
            If True, the bolometric flux is adjusted by
            ``F_out = F_in * (D_in / D_out) ** 2``, where ``D_in`` and
            ``D_out`` are current and target luminosity distances,
            respectively. ``D_in`` is self.dist. If self.dist is ``None``,
            the distance is calculated from the current redshift and
            given cosmology.
        dist : float, optional
            Output distance in Mpc. Used to adjust bolometric flux if
            ``adjust_flux`` is ``True``. Default is ``None`` which means
            that the distance is calculated from the redshift and the
            cosmology.
        cosmo : `~astropy.cosmology.Cosmology` instance, optional
            The cosmology used to estimate distances if dist is not given.
            Default is ``None``, which results in using the default
            cosmology.

        Returns
        -------
        spec : Spectrum object
            A new spectrum object at redshift z.
        """

        if self._z is None:
            raise ValueError('Must set current redshift in order to redshift'
                             ' spectrum')

        if self._dunit.physical_type == u.m.physical_type:
            factor = (1. + z) / (1. + self._z)
        elif self._dunit.physical_type == u.Hz.physical_type:
            factor = (1. + self._z) / (1. + z)
        else:
            raise ValueError('dispersion must be in wavelength or frequency')

        d = self._disp * factor
        f = self._flux / factor
        if self._error is not None:
            e = self._error / factor
        else:
            e = None

        if adjust_flux:
            if self._dist is None and self._z == 0.:
                raise ValueError("When current redshift is 0 and adjust_flux "
                                 "is requested, current distance must be "
                                 "defined")
            if dist is None and z == 0.:
                raise ValueError("When redshift is 0 and adjust_flux "
                                 "is requested, dist must be defined")
            if cosmo is None:
                cosmo = cosmology.get_current()

            if self._dist is None:
                dist_in = cosmo.luminosity_distance(self._z)
            else:
                dist_in = self._dist

            if dist is None: dist = cosmo.luminosity_distance(z)

            if dist_in <= 0. or dist <= 0.:
                raise ValueError("Distances must be greater than 0.")

            # Adjust the flux
            factor = (dist_in / dist) ** 2
            f *= factor
            if e is not None: e *= factor

        return Spectrum(d, f, error=e, z=z, dist=dist, meta=self.meta,
                        unit=self._unit, dunit=self._dunit)


class MagSystem(object):
    """An abstract base class for magnitude systems."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, refmags=None, name=None):
        self._zpbandflux = {}
        self._name = name
        self._refmags = refmags

    @abc.abstractmethod
    def _refspectrum_bandflux(self, band):
        """Flux of the fundamental spectrophotometric standard."""
        pass

    @property
    def name(self):
        """Name of magnitude system."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


    @property
    def refmags(self):
        """Dictionary of magnitude refmags for particular bands."""
        return copy.copy(self._refmags)

    @refmags.setter
    def refmags(self, new_refmags):
        if new_refmags is None:
            self._refmags = None
            return

        if not isinstance(new_refmags, dict):
            raise ValueError("refmags must be dict-like or None")

        self._refmags = OrderedDict()
        for band, offset in new_refmags:
            self._refmags[get_bandpass(band)] = offset

    def zpbandflux(self, band):
        """Flux of an object with magnitude zero in the given bandpass.

        Parameters
        ----------
        bandpass : `~sncosmo.spectral.Bandpass` or str

        Returns
        -------
        bandflux : float
            Flux in photons / s / cm^2.
        """

        band = get_bandpass(band)
        try:
            return self._zpbandflux[band]
        except KeyError:
            pass

        bandflux = self._refspectrum_bandflux(band)
        if self._refmags is not None:
            if band not in self._refmags:
                raise Exception('Band refmags not defined')
            bandflux *= 10 ** (0.4 * self._refmags[band])
    
        self._zpbandflux[band] = bandflux
        return bandflux

    def band_flux_to_mag(self, flux, band):
        """Convert flux (photons / s / cm^2) to magnitude."""
        return -2.5 * math.log10(flux / self.zpbandflux(band))

    def band_mag_to_flux(self, mag, band):
        """Convert magnitude to flux in photons / s / cm^2"""
        return self.zpbandflux(band) * 10.**(-0.4 * mag)


class SpectralMagSystem(MagSystem):
    """A magnitude system defined by a fundamental spectrophotometric
    standard.

    Parameters
    ----------
    refspectrum : `sncosmo.Spectrum`
        The spectrum of the fundamental spectrophotometric standard.
    refmags : dict, optional
        The magnitudes (in this magnitude system) of the fundamental
        spectrophotometric standard in the given bandpasses. If `None`
        (the default), it is assumed that the fundamental standard has
        magnitude 0 in all bandpasses. Dictionary keys can either be
        `sncosmo.Bandpass` objects or string identifiers
        thereof. Dictionary values are floats.
    """

    def __init__(self, refspectrum, refmags=None, name=None):
        super(SpectralMagSystem, self).__init__(refmags, name)
        self._refspectrum = refspectrum

    def _refspectrum_bandflux(self, band):
        return self._refspectrum.bandflux(band)


class ABMagSystem(MagSystem):
    """Magnitude system where a source with F_nu = 3631 Jansky at all
    frequencies has magnitude 0 in all bands."""
    
    def _refspectrum_bandflux(self, band):
        bdisp, btrans = band.to_unit(u.Hz)

        # AB spectrum is 3631 x 10^{-23} erg/s/cm^2/Hz
        # Get spectral values in photons/cm^2/s/Hz at bandpass dispersions
        # by dividing by (h \nu).
        f = 3631.e-23 / const.h.cgs.value / bdisp

        binw = np.gradient(bdisp)
        return np.sum(f * btrans * binw)


#    Examples
#    --------
#    >>> chft = LocalMagSystem(bd17spectrum,
#    ...                       {'megacamu': 9.7688, 'megacamg': 9.6906,
#    ...                        'megacamr': 9.2183, 'megacami': 8.9142})
#    >>> chft.zp_cflux('megacamu') # flux an object 9.7688 mag brighter than bd17
#    >>> chft.zp_cflux('desg') # Exception
#    """






import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import Grid
from sncosmo import registry
from sncosmo import Bandpass, get_bandpass


bandpass_meta = registry.get_loaders_metadata(Bandpass)
filtersets = []
for m in bandpass_meta:
    if m['filterset'] not in filtersets: filtersets.append(m['filterset'])

fig = plt.figure(figsize=(9., 3. * len(filtersets)))
grid = Grid(fig, rect=111, nrows_ncols=(len(filtersets), 1),
            axes_pad=0.25, label_mode='L')

for ax, filterset in zip(grid, filtersets):

    for m in bandpass_meta:
        if m['filterset'] != filterset: continue
        b = get_bandpass(m['name'])
        ax.plot(b.disp, b.trans, label=m['name'])
    ax.set_xlabel('Angstroms')
    ax.set_ylabel('Transmission')
    ax.legend(loc='upper right')

xmin, xmax = ax.get_xlim()
ax.set_xlim(right=(xmax + 1000.))  # make room for legend
plt.tight_layout()
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division

import numpy as np

from scipy.optimize import fmin_l_bfgs_b
from astropy.utils import OrderedDict

from spectral import get_magsystem

__all__ = ['fit_model']

def normalized_flux(data, zp=25., magsys='ab'):
    """Return flux values normalized to a common zeropoint and magnitude
    system."""

    magsys = get_magsystem(magsys)
    flux = np.empty(len(data['flux']), dtype=np.float)
    fluxerr = np.empty(len(data['flux']), dtype=np.float)

    for i in range(len(data['flux'])):
        ms = get_magsystem(data['zpsys'][i])
        factor = (ms.zpbandflux(data['band'][i]) /
                  magsys.zpbandflux(data['band'][i]) *
                  10.**(0.4 * (zp - data['zp'][i])))
        flux[i] = data['flux'][i] * factor
        fluxerr[i] = data['fluxerr'][i] * factor

    return flux, fluxerr

def guess_parvals(data, model, parnames=['t0', 'fscale']):

    nflux, nfluxerr = normalized_flux(data, zp=25., magsys='ab')
    bandt0 = []
    bandfluxscale = []
    for band in np.unique(data['band']):
        idx = data['band'] == band
        flux = nflux[idx]
        time = data['time'][idx]
        weights = flux ** 2 / nfluxerr[idx]
        topn = min(len(weights) // 2, 3)
        if topn == 0: continue
        topnidx = np.argsort(weights)[-topn:]
        bandt0.append(np.average(time[topnidx], weights=weights[topnidx]))
        maxdataflux = np.average(flux[topnidx], weights=weights[topnidx])
        maxmodelflux = max(model.bandflux(band, zp=25., zpsys='ab'))
        bandfluxscale.append(maxdataflux / maxmodelflux) 

    t0 = sum(bandt0) / len(bandt0)
    fscale = (sum(bandfluxscale) / len(bandfluxscale) *
                  model.params['fscale'])

    result = {}
    if 't0' in parnames: result['t0'] = t0
    if 'fscale' in parnames: result['fscale'] = fscale
    return result
                       
def fit_model(model, data, parnames, bounds=None, params_start=None,
              t0range=20., verbose=False, include_model_error=False):
    """Fit model parameters to data by minimizing chi^2.

    Parameters
    ----------
    model : `~sncosmo.Model`
        The model to fit.
    data : `~numpy.ndarray` or `dict`
        Table containing columns 'date', 'band', 'flux', 'fluxerr', 'zp',
        'zpsys'.
    parnames : list
        Model parameters to vary in the fit.
    bounds : `dict`, optional
        Bounded range for each parameter. Keys should be parameter names,
        values are tuples. If a bound is not given for some parameter,
        the parameter is unbounded. The exception is ``t0``, which has a
        default bound of  ``(initial guess) +/- t0range``.
    params_start : `dict`, optional
        If given, use these initial parameters in fit. Default is to use
        current model parameters.
    t0range : float, optional
        Bounds for t0 (if varied in fit and not given in `bounds`).
        Default is 20.
    verbose : bool, optional
        Print minimization info to the screen.

    Returns
    -------
    min_chisq : float
        Value of Chi^2 for fitted model parameters. The model's parameters
        are set to the best-fit parameters.

    Notes
    -----
    Uses scipy's L-BFGS-B bounded minimization algorithm.
    """

    # Check that if z is going to be fit, it is bounded.
    if 'z' in parnames and (bounds is None or 'z' not in bounds):
        raise ValueError('z must be bounded if fit.')

    # Check redshift range to see which bands we can use in the fit.
    bands = np.unique(data['band'])
    if 'z' not in parnames:
        valid = model.bandoverlap(data['band'])
    else:
        valid = model.bandoverlap(data['band'], z=bounds['z'])
        valid = np.all(valid, axis=1)
    if not np.all(valid):
        print "WARNING: dropping following bands from data:"
        print np.unique(data['band'][np.invert(valid)])
        data = {'time': data['time'][valid],
                'band': data['band'][valid],
                'flux': data['flux'][valid],
                'fluxerr': data['fluxerr'][valid],
                'zp': data['zp'][valid],
                'zpsys': data['zpsys'][valid]}

    # If we're fitting redshift and it is bounded, set initial value
    if 'z' in parnames:
        model.set(z=(sum(bounds['z']) / 2.))

    # Get initial guesses
    parvals0 = []
    guesses = guess_parvals(data, model, parnames=['t0', 'fscale'])
    current = model.params
    for name in parnames:
        if name in params_start:
            parvals0.append(params_start[name])
        elif name in guesses:
            parvals0.append(guesses[name])
        else:
            parvals0.append(current[name])

    if bounds is None:
        bounds = {}
    bounds_list = []
    for name in parnames:
        if name in bounds:
            bounds_list.append(bounds[name])
        elif name == 't0':
            bounds_list.append((guesses['t0'] - t0range,
                                guesses['t0'] + t0range))
        else:
            bounds_list.append((None, None))

    if verbose:
        print "starting point:"
        for name, val, bound in zip(parnames, parvals0, bounds_list):
            print "   ", name, val, bound

    # scale fscale, for numerical precision reasons.
    fscale_factor = 1.
    if 'fscale' in parnames:
        i = parnames.index('fscale')
        fscale_factor = parvals0[i]
        parvals0[i] = 1.
        if 'fscale' in bounds:
            bounds_list[i] = (bounds_list[i][0] / fscale_factor,
                              bounds_list[i][1] / fscale_factor)

    def chi2(parvals):
        params = dict(zip(parnames, parvals))
        if 'fscale' in params:
            params['fscale'] *= fscale_factor
        model.set(**params)
        if include_model_error:
            modelflux, modelfluxerr = model.bandflux(
                data['band'], data['time'], zp=data['zp'],
                zpsys=data['zpsys'], include_error=True)
            return np.sum((data['flux'] - modelflux) ** 2 /
                          (modelfluxerr ** 2 + data['fluxerr'] ** 2))
        else:
            modelflux = model.bandflux(data['band'], data['time'],
                                       zp=data['zp'], zpsys=data['zpsys'])
            return np.sum(((data['flux'] - modelflux) / data['fluxerr']) ** 2)

    parvals, fval, d = fmin_l_bfgs_b(chi2, parvals0, bounds=bounds_list,
                                     approx_grad=True, iprint=(verbose - 1))
    params = dict(zip(parnames, parvals))
    if 'fscale' in params:
        params['fscale'] *= fscale_factor
    model.set(**params)
    return fval



#!/usr/bin/env python

from matplotlib import cm
import matplotlib.pyplot as plt
import sncosmo
cmap = cm.get_cmap('gist_rainbow')
model = sncosmo.get_model('hsiao')

disp = model.disp()
flux = model.flux(0.)

dmin = 2600.
dmax = 7400.
idx = (disp > dmin) & (disp < dmax)
disp = disp[idx]
flux = flux[idx]

cscale = (dmax - disp[1:]) / (dmax - dmin)
colors = cmap(cscale)

fig = plt.figure(figsize=(1., 1.), facecolor=(0., 0., 0., 0.))
ax = plt.axes([0., 0., 1., 1.], axisbg=(0.,0., 0., 0.))

for i in range(len(cscale)):
    d0, d1 = disp[i], disp[i+1]
    f0, f1 = flux[i], flux[i+1]
    plt.fill([d0, d0, d1, d1], [0., f0, f1, 0.], color=colors[i])
plt.xlim(dmin, dmax)
plt.ylim(0., 1.05*max(flux))
plt.axis('off')
plt.savefig('logo.png', dpi=500., transparent=True)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Something about the registry."""

from astropy.utils import OrderedDict

__all__ = ['register_loader', 'register', 'retrieve', 'get_loaders_metadata']

_loaders = OrderedDict()
_instances = OrderedDict()

def register_loader(data_class, name, function, function_args=None,
                    version=None, force=False, **meta):
    """Register a data reading function.

    Parameters
    ----------
    data_class : classobj
        The class of the object that the loader returns.
    name : str
        The data identifier.
    function : function
        The function to read in the data. Should accept a name and version
        keyword argument.
    function_args : list, optional
        Arguments to pass to the function. Default is an empty list.
    version : str, optional
        Sub-version of name, if desired. Use formats such as ``'1'``,
        ``'1.0'``, ``'1.0.0'``, etc. Default is `None`. 
    force : bool, optional
        Whether to override any existing function if already present.
    **meta : optional
        Any additional keyword arguments are assumed to be metadata and
        are saved as a dictionary describing this loader.
    """

    if function_args is None:
        function_args = []

    if version is None:
        key = (data_class, name)
    else:
        key = (data_class, name, version)

    if not key in _loaders or force:
        _loaders[key] = function, function_args, meta
    else:
        if version is not None:
            versionstr = " (version='{0:s}')".format(version)
        else:
            versionstr = ""
        raise Exception("Loader for {0:s} named '{1:s}'{2:s} is already "
                        "defined. Use force=True to override."
                        .format(data_class.__name__, name, versionstr))


def register(instance, name=None, data_class=None, force=False):
    """Register a class instance.

    Parameters
    ----------
    instance : object
        The object to be registered.
    name : str, optional
        Identifier. If `None`, the name is taken from the `name` attribute
        of the instance, if it exists and is a string.
    data_class : classobj, optional
        If given, the instance is registered as an instance of this class
        rather the the class of the instance itself. Use this for registering
        subclasses when you wish them to be accessible from their superclass.
    force : bool, optional
        Whether to override any existing instance of the same name. Note: this
        may not play well with versioned instances.
    """

    if name is None:
        try:
            name = instance.name
        except AttributeError:
            raise ValueError("name not given and instance has no 'name' "
                             "attribute")
        if not isinstance(name, basestring):
            raise ValueError("name attribute of {0!r:s} is not a string.")

    if data_class is None:
        data_class = instance.__class__
        
    key = (data_class, name)
    
    already_present = (key in _instances) or (key in _loaders)
    if not already_present or force:
        _loaders[key] = instance
    else:
        raise Exception("{0:s} named {1:s} already in registry. Use force=True"
                        " to override.".format(dat_class.__name__, name))

def retrieve(data_class, name, version=None):
    """Retrieve a class instance from a registered identifier.

    Parameters
    ----------
    data_class : classobj
        The class of the object requested.
    name : str
        Identifier of the specific instance.
    version : str
        Sub-identifier. If `None`, default to highest or only version.

    Returns
    -------
    instance : data_class (or subclass thereof)

    Notes
    -----
    **Precedence** The following are tried in this order:

    1. If `name` is already an instance of `data_class`
       (rather than a string), it is immediately returned.
    2. If (`data_class`, `name`) is already in the registry,
       that instance is returned.
    3. If there is a loader defined for (`data_class`, `name`), it is used
       to create an instance, save it to the registry and return it.
    4. An Exception is raised listing the available registered names for
       the requested data class.

    **Versioning** There is support for multiple versions of data for the
    same `name`.

    1. If `version` is specified, the registry and its loaders are
       searched for (`data_class`, `name`, `version`).
    2. If `version` is not specified but there are registered loaders for
       (`data_class`, `name`, `version`), the latest version is used,
       and both (`data_class`, `name`) and (`data_class`, `name`,
       `version`) are saved to the registry. "Latest" is defined by
       string comparision.

    """

    if isinstance(name, data_class): return name
    if version is None:
        key = (data_class, name)
    else:
        key = (data_class, name, version)

    try:
        return _instances[key]
    except KeyError:
        pass

    if key in _loaders:
        function, function_args, meta = _loaders[key]
        if version is None:
            _instances[key] = function(*function_args, name=name)
        else:
            _instances[key] = function(*function_args, name=name, 
                                        version=version)
            
        return _instances[key]

    if version is None:
        latest_version = None
        for regkey in _loaders.keys():
            if (key == regkey[0:2] and
                (regkey[2] > latest_version or latest_version is None)):
                latest_version = regkey[2]
        if latest_version is not None:
            regkey = (key[0], key[1], latest_version)
            function, function_args, meta = _loaders[regkey]
            _instances[regkey] = function(
                *function_args, name=name, version=latest_version)
            _instances[key] = _instances[regkey]
            return _instances[key]

    # At this point we will raise an exception.
    registered_names = [regkey[1] for regkey in _loaders.keys()
                        if regkey[0] is data_class]

    if version is None or name not in registered_names:
        raise Exception(
            "No {0:s} named '{1:s}' in registry. Registered names: '{2:s}'"
            .format(data_class.__name__, name, "', '".join(registered_names)))

    registered_versions = [
        regkey[2] for regkey in _loaders.keys() if key[0:2] == regkey[0:2]]
    raise Exception(
        "No {0:s} named '{1:s}' with version='{2:s}' in registry. Registered"
        " versions: '{3:s}'".format(data_class.__name__, name, version,
                                    "', '".join(registered_versions)))

def get_loaders_metadata(data_class):
    """Return the metadata of all registered loaders for a given class.
    
    Parameters
    ----------
    data_class : classobj

    Returns
    -------
    loadermeta : list of dict
        Each item in the list is a dictionary containing a 'name'
        keyword, a 'version' keyword (if applicable), and the metadata
        keywords for the given loader.
    """

    loaders_metadata = []
    for lkey, loader in _loaders.iteritems():
        if lkey[0] is not data_class: continue
        m = {'name': lkey[1]}
        if len(lkey) > 2: m['version'] = lkey[2]
        m.update(loader[2])
        loaders_metadata.append(m)
    return loaders_metadata



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions for supernova light curve I/O"""

# This module is designed to be self-contained, with only a dependency on
# numpy

# Try to get OrderedDict from collections, then from astropy, then just use
# the regular built-in dictionary.
try:
    from collections import OrderedDict as odict
except ImportError:
    try:
        from astropy.utils import OrderedDict as odict
    except ImportError:
        odict = dict

HAS_NUMPY = True
try:
    import numpy as np
except ImportError:
    HAS_NUMPY = False

__all__ = ['readlc', 'writelc']

def _cast_str(s):
    try:
        return int(s)
    except:
        try:
            return float(s)
        except:
            return s.strip()

def dict_to_array(d):
    """Convert a dictionary of lists (of equal length) to a structured
    numpy.ndarray"""
    import numpy as np

    # first convert all lists to 1-d arrays, in order to let numpy
    # figure out the necessary size of the string arrays.
    for key in d: 
        d[key] = np.array(d[key])

    # Determine dtype of output array.
    dtypelist = []
    for key in d:
        dtypelist.append((key, d[key].dtype))
    
    # Initialize ndarray and then fill it.
    firstkey = d.keys()[0]
    col_len = len(d[firstkey])
    result = np.empty(col_len, dtype=dtypelist)
    for key in d:
        result[key] = d[key]

    return result

# ------------------------------- Readers ------------------------------------
def _read_csv(f, **kwargs):

    delim = kwargs.get('delim', None)
    metachar = kwargs.get('metachar', '@')
    commentchar = kwargs.get('commentchar', '#')

    meta = odict()
    colnames = []
    cols = []
    readingdata = False
    for line in f:
        
        # strip leading & trailing whitespace, newline, and comments
        line = line.strip()
        pos = line.find(commentchar)
        if pos > -1:
            line = line[:pos]
        if len(line) == 0: continue

        if not readingdata:
            # Read metadata
            if line[0] == metachar:
                pos = line.find(' ')  # Find first space.
                if pos in [-1, 1]:  # Space must exist and key must exist.
                    raise ValueError('Incorrectly formatted metadata line: '+
                                     line)
                meta[line[1:pos]] = _cast_str(line[pos:])
                continue

            # Read header line
            for item in line.split(delim):
                colnames.append(item.strip())
                cols.append([])
            readingdata = True
            continue

        # Now we're reading data
        items = line.split(delim)
        for col, item in zip(cols, items):
            col.append(_cast_str(item))

    data = odict(zip(colnames, cols))
    return meta, data

# SALT2 conversions upon reading:
# Names are converted to lowercase, then the following lookup table is used.
SALT2KEY_TO_KEY = {'redshift': 'z',
                   'z_heliocentric': 'z_helio',
                   'filter': 'band'}
def _salt2_rename_keys(d):
    newd = odict()
    for key, val in d.iteritems():
        key = key.lower()
        if key in SALT2KEY_TO_KEY:
            key = SALT2KEY_TO_KEY[key]
        newd[key] = val
    return newd

def _read_salt2(f, **kwargs):
    """Read a new-style SALT2 file.
    
    Such a file has metadata on lines starting with '@' and column names
    on lines starting with '#' and containing a ':' after the column name.
    There is optionally a line containing '#end' before the start of data.
    """

    raw = kwargs.get('raw', False)

    meta = odict()
    colnames = []
    cols = []
    readingdata = False
    for line in f:
        
        # strip leading & trailing whitespace & newline
        line = line.strip()
        if len(line) == 0: continue

        if not readingdata:
            # Read metadata
            if line[0] == '@':
                pos = line.find(' ')  # Find first space.
                if pos in [-1, 1]:  # Space must exist and key must exist.
                    raise ValueError('Incorrectly formatted metadata line: '+
                                     line)
                meta[line[1:pos]] = _cast_str(line[pos:])
                continue

            # Read header line
            if line[0] == '#':
                pos = line.find(':')
                if pos in [-1, 1]: continue  # comment line
                colname = line[1:pos].strip()
                if colname == 'end': continue 
                colnames.append(colname)
                cols.append([])
                continue

            # If the first non-whitespace character is not '@' or '#',
            # assume the line is the first data line.
            readingdata = True

        # strip comments
        pos = line.find('#')
        if pos > -1:
            line = line[:pos]
        if len(line) == 0: continue

        # Now we're reading data
        items = line.split()
        for col, item in zip(cols, items):
            col.append(_cast_str(item))

    data = odict(zip(colnames, cols))

    # convert keys, if requested
    if not raw:
        meta = _salt2_rename_keys(meta)
        data = _salt2_rename_keys(data)

    return meta, data

READERS = {'csv': _read_csv,
           'salt2': _read_salt2}

def readlc(fname, fmt='csv', array=True, **kwargs):
    """Read light curve data.

    Parameters
    ----------
    fname : str
        Filename 
    fmt : {'csv', 'salt2', 'snana'}, optional
        Format of file. Default is 'csv'. 'salt2' is the new format available
        in snfit version >= 2.3.0.
    array : bool, optional
        If True (default), returned data is a numpy array, Otherwise, data
        is a dictionary of lists (representing columns).
    delim : str, optional
        **[csv only]** Used to split entries on a line. Default is `None`.
        Extra whitespace is ignored.
    metachar : str, optional
        **[csv only]** Lines whose first non-whitespace character is `metachar`
        are treated as metadata lines, where the key and value are split on
        the first whitespace. Default is '@'
    commentchar : str, optional
        **[csv only]** One-character string indicating a comment. Default is
        '#'.
    raw : bool, optional
        **[salt2 only]** By default, the SALT2 reader converts all column names
        and metadata keys to lowercase and renames a few of them. Set to True
        to override this. Default is False.

    Returns
    -------
    meta : dict
        A (possibly empty) dictionary of metadata in the file.
    data : `~numpy.ndarray` or dict
        Data.
    """

    if fmt not in READERS:
        raise ValueError("Reader not defined for format '{}'. Options: "
                         .format(fmt) + ", ".join(READERS.keys()))

    with open(fname, 'rb') as f:
        meta, data = READERS[fmt](f, **kwargs)

    if array:
        if HAS_NUMPY:
            data = dict_to_array(data)
        else:
            raise ValueError('numpy required for array output. Use '
                             'array=False for dictionary output.')
    return meta, data


# ------------------------------- Writers -----------------------------------#
def _write_csv(f, data, meta, **kwargs):
    
    delim = kwargs.get('delim', ' ')
    metachar = kwargs.get('metachar', '@')
    
    if meta is not None:
        for key, val in meta.iteritems():
            f.write('{}{}{}{}\n'.format(metachar, key, delim, str(val)))

    if hasattr(data, 'dtype'):
        keys = data.dtype.names
        length = len(data)
    else:
        keys = data.keys()
        length = len(data[keys[0]])
    
    f.write(delim.join(keys))
    f.write('\n')
    for i in range(length):
        f.write(delim.join([str(data[key][i]) for key in keys]))
        f.write('\n')

KEY_TO_SALT2KEY_META = {
    'Z': 'REDSHIFT',              # Not sure if this is used.
    'Z_HELIOCENTRIC': 'Z_HELIO',
    'MAGSYS': 'MagSys',
    'Z_SOURCE': 'z_source'}
KEY_TO_SALT2KEY_COLUMN = {
    'Mjd': 'Date',
    'Time': 'Date',
    'Flux': 'FluxPsf',
    'Fluxpsf': 'FluxPsf',
    'Fluxerr': 'FluxPsferr',
    'Fluxpsferr': 'FluxPsferr',
    'Airmass': 'AirMass',
    'Zp': 'ZP',
    'Zpsys': 'MagSys',
    'Magsys': 'MagSys',
    'Band': 'Filter'}

def _write_salt2(f, data, meta, **kwargs):

    raw = kwargs.get('raw', False)
    pedantic = kwargs.get('pedantic', True)
    
    if meta is not None:
        for key, val in meta.iteritems():
            if not raw:
                key = key.upper()
                key = KEY_TO_SALT2KEY_META.get(key, key)
            f.write('@{} {}\n'.format(key, str(val)))

    if hasattr(data, 'dtype'):
        keys = data.dtype.names
        length = len(data)
    else:
        keys = data.keys()
        length = len(data[keys[0]])

    # Write column names
    keys_as_written = []
    for key in keys:
        if not raw:
            key = key.capitalize()
            key = KEY_TO_SALT2KEY_COLUMN.get(key, key)
        f.write('#{} :\n'.format(key))
        keys_as_written.append(key)
    f.write('#end :\n')

    # Check that necessary fields exist
    if pedantic:
        if not ('Filter' in keys_as_written and 'MagSys' in keys_as_written):
            raise ValueError('photometry data missing required some fields '
                             ': Filter, MagSys')

    # Write the data itself
    for i in range(length):
        f.write(' '.join([str(data[key][i]) for key in keys]))
        f.write('\n')


KEY_TO_SNANAKEY_COLUMN = {
    'TIME': 'MJD',
    'DATE': 'MJD',
    'FILTER': 'FLT',
    'BAND': 'FLT',
    'FLUX': 'FLUXCAL',
    'FLUXERR': 'FLUXCALERR',
    'ZP': 'ZPT',
    'ZEROPOINT': 'ZPT'
    }
KEY_TO_SNANAKEY_META = {
    'DEC': 'DECL'
    }
SNANA_REQUIRED_META = ['RA', 'DECL', 'SURVEY', 'FILTERS', 'MWEBV']
SNANA_REQUIRED_COLUMN = ['MJD', 'FLT', 'FLUXCAL', 'FLUXCALERR', 'ZPT']
def _write_snana(f, data, meta, **kwargs):

    raw = kwargs.get('raw', False)
    pedantic = kwargs.get('pedantic', True)

    # Write metadata
    keys_as_written = []
    if meta is not None:
        for key, val in meta.iteritems():
            if not raw:
                key = key.upper()
                key = KEY_TO_SNANAKEY_META.get(key, key)
            f.write('{}: {}\n'.format(key, str(val)))
            keys_as_written.append(key)

    # Check that necessary metadata was written
    if pedantic:
        for key in SNANA_REQUIRED_META:
            if key not in keys_as_written:
                raise ValueError('Missing required metadata kw: ' + key)
    
    # Get column names and data length
    if hasattr(data, 'dtype'):
        keys = data.dtype.names
        length = len(data)
    else:
        keys = data.keys()
        length = len(data[keys[0]])
    
    # Convert column names
    keys_to_write = []
    for key in keys:
        if not raw:
            key = key.upper()
            key = KEY_TO_SNANAKEY_COLUMN.get(key, key)
        keys_to_write.append(key)

    # Check that necessary column names are included
    if pedantic:
        for key in SNANA_REQUIRED_COLUMN:
            if key not in keys_to_write:
                raise ValueError('Missing required column name: ' + key)

    # Write the header
    f.write('\n'
            '# ==========================================\n'
            '# TERSE LIGHT CURVE OUTPUT:\n'
            '#\n'
            'NOBS: {:d}\n'
            'NVAR: {:d}\n'
            'VARLIST: {}\n'
            .format(length, len(keys), ' '.join(keys_to_write)))

    # Write data
    for i in range(length):
        f.write('OBS: ')
        f.write(' '.join([str(data[key][i]) for key in keys]))
        f.write('\n')

WRITERS = {'csv': _write_csv,
           'salt2': _write_salt2,
           'snana': _write_snana}

def writelc(data, fname, meta=None, fmt='csv', **kwargs):
    """Write light curve data.

    Parameters
    ----------
    data : `~numpy.ndarray` or dict
        Data.
    fname : str
        Filename.
    meta : dict, optional
        A (possibly empty) dictionary of metadata. Default is None.
    fmt : {'csv', 'salt2', 'snana'}, optional
        Format of file. Default is 'csv'. 'salt2' is the new format available
        in snfit version >= 2.3.0.
    delim : str, optional
        **[csv only]** Character used to separate entries on a line.
        Default is ' '.
    metachar : str, optional
        **[csv only]** Metadata designator. Default is '@'.
    raw : bool, optional
        **[salt2, snana]** By default, the SALT2 and SNANA writers rename
        some metadata keys and column names in order to comply with what
        snfit and SNANA expect. Set to True to override this.
        Default is False.
    pedantic : bool, optional
        **[salt2, snana]** If True, check that output column names and header
        keys comply with expected formatting, and raise a ValueError if not.
        It is probably a good idea to set to False when raw is True.
        Default is True. 
    """

    if fmt not in WRITERS:
        raise ValueError("Writer not defined for format '{}'. Options: "
                         .format(fmt) + ", ".join(WRITERS.keys()))
    with open(fname, 'w') as f:
        WRITERS[fmt](f, data, meta, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Reader functions for initializing built-in data."""

import string
import tarfile
from os.path import join

import numpy as np
from astropy.io import fits
from astropy import wcs
from astropy.utils.data import (download_file, get_pkg_data_filename,
                                get_readable_fileobj)

from .. import registry
from .. import Model, TimeSeriesModel, StretchModel, SALT2Model
from .. import io

# ------------------------------------------------------------------------
# Bandflux relative errors

def set_bandfluxerror_sn1a(model):
    dmin, dmax = model.disp(modelframe=True)[[0, -1]]
    disp = np.linspace(dmin, dmax, (dmax-dmin)/100. + 1)
    phase = model.times(modelframe=True)
    relphase = np.abs(phase - model.refphase)
    relerror = np.empty(relphase.shape, dtype=np.float)
    idx = relphase < 20.
    relerror[idx] = 0.08 + 0.04 * relphase[idx] / 20.
    idx = np.invert(idx)
    relerror[idx] = 0.12 + 0.08 * (relphase[idx] - 20.) / 60.
    relerror = relerror[:, np.newaxis]
    relerror = relerror * (((disp < 4000.) | (disp > 8300.)) + 1.)
    model.set_bandflux_relative_error(phase, disp, relerror)

def set_bandfluxerror_sncc(model):
    dmin, dmax = model.disp(modelframe=True)[[0, -1]]
    disp = np.linspace(dmin, dmax, (dmax-dmin)/100. + 1)
    phase = model.times(modelframe=True)
    relphase = phase - model.refphase
    relerror = (0.08 + 0.08 * np.abs(relphase) / 60.)[:, np.newaxis]
    relerror = relerror * (((disp < 4000.) | (disp > 8300.)) + 1.)
    model.set_bandflux_relative_error(phase, disp, relerror)

# ------------------------------------------------------------------------
# Nugent models

def load_timeseries_ascii_sn1a(remote_url, name=None, version=None):
    with get_readable_fileobj(remote_url, cache=True) as f:
        phases, wavelengths, flux = io.read_griddata(f)
    model = StretchModel(phases, wavelengths, flux,
                         name=name, version=version)
    set_bandfluxerror_sn1a(model)
    return model

def load_timeseries_ascii_sncc(remote_url, name=None, version=None):
    with get_readable_fileobj(remote_url, cache=True) as f:
        phases, wavelengths, flux = io.read_griddata(f)
    model = TimeSeriesModel(phases, wavelengths, flux,
                            name=name, version=version)
    set_bandfluxerror_sncc(model)
    return model

# ------------------------------------------------------------------------
# Nugent models

nugent_baseurl = 'http://supernova.lbl.gov/~nugent/templates/'
nugent_website = 'http://supernova.lbl.gov/~nugent/nugent_templates.html'
nugent_subclass_1a = '`~sncosmo.StretchModel`'
nugent_subclass_cc = '`~sncosmo.TimeSeriesModel`'


registry.register_loader(
    Model, 'nugent-sn1a', load_timeseries_ascii_sn1a, 
    [nugent_baseurl + 'sn1a_flux.v1.2.dat.gz'],
    version='1.2', url=nugent_website, type='SN Ia',
    subclass=nugent_subclass_1a,
    reference=('N02', 'Nugent, Kim & Permutter 2002 '
               '<http://adsabs.harvard.edu/abs/2002PASP..114..803N>'))

registry.register_loader(
    Model, 'nugent-sn91t', load_timeseries_ascii_sn1a, 
    [nugent_baseurl + 'sn91t_flux.v1.1.dat.gz'],
    version='1.1', url=nugent_website, type='SN Ia',
    subclass=nugent_subclass_1a,
    reference=('S04', 'Stern, et al. 2004 '
               '<http://adsabs.harvard.edu/abs/2004ApJ...612..690S>'))

registry.register_loader(
    Model, 'nugent-sn91bg', load_timeseries_ascii_sn1a, 
    [nugent_baseurl + 'sn91bg_flux.v1.1.dat.gz'],
    version='1.1', url=nugent_website, type='SN Ia',
    subclass=nugent_subclass_1a,
    reference=('N02', 'Nugent, Kim & Permutter 2002 '
               '<http://adsabs.harvard.edu/abs/2002PASP..114..803N>'))

registry.register_loader(
    Model, 'nugent-sn1bc', load_timeseries_ascii_sncc, 
    [nugent_baseurl + 'sn1bc_flux.v1.1.dat.gz'],
    version='1.1', url=nugent_website, type='SN Ib/c',
    subclass=nugent_subclass_cc,
    reference=('L05', 'Levan et al. 2005 '
               '<http://adsabs.harvard.edu/abs/2005ApJ...624..880L>'))

registry.register_loader(
    Model, 'nugent-hyper', load_timeseries_ascii_sncc, 
    [nugent_baseurl + 'hyper_flux.v1.2.dat.gz'],
    version='1.2', url=nugent_website, type='SN Ib/c',
    subclass=nugent_subclass_cc,
    reference=('L05', 'Levan et al. 2005 '
               '<http://adsabs.harvard.edu/abs/2005ApJ...624..880L>'))

registry.register_loader(
    Model, 'nugent-sn2p', load_timeseries_ascii_sncc, 
    [nugent_baseurl + 'sn2p_flux.v1.2.dat.gz'],
    version='1.2', url=nugent_website, type='SN IIP',
    subclass=nugent_subclass_cc,
    reference=('G99', 'Gilliland, Nugent & Phillips 1999 '
               '<http://adsabs.harvard.edu/abs/1999ApJ...521...30G>'))

registry.register_loader(
    Model, 'nugent-sn2l', load_timeseries_ascii_sncc, 
    [nugent_baseurl + 'sn2l_flux.v1.2.dat.gz'],
    version='1.2', url=nugent_website, type='SN IIL',
    subclass=nugent_subclass_cc,
    reference=('G99', 'Gilliland, Nugent & Phillips 1999 '
               '<http://adsabs.harvard.edu/abs/1999ApJ...521...30G>'))

registry.register_loader(
    Model, 'nugent-sn2n', load_timeseries_ascii_sncc, 
    [nugent_baseurl + 'sn2n_flux.v2.1.dat.gz'],
    version='2.1', url=nugent_website, type='SN IIn',
    subclass=nugent_subclass_cc,
    reference=('G99', 'Gilliland, Nugent & Phillips 1999 '
               '<http://adsabs.harvard.edu/abs/1999ApJ...521...30G>'))

del nugent_website
del nugent_baseurl
del nugent_subclass_1a
del nugent_subclass_cc

# -----------------------------------------------------------------------
# Sako et al 2011 models

s11_baseurl = 'http://kbarbary.github.com/data/models/'
s11_ref = ('S11', 'Sako et al. 2011 '
           '<http://adsabs.harvard.edu/abs/2011ApJ...738..162S>')
s11_website = 'http://sdssdp62.fnal.gov/sdsssn/SNANA-PUBLIC/'
s11_subclass = '`~sncosmo.TimeSeriesModel`'
s11_note = "extracted from SNANA's SNDATA_ROOT on 29 March 2013."

registry.register_loader(
    Model, 's11-2004hx', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-000018.SED'], version='1.0', url=s11_website,
    type='SN IIL/P', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2005lc', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-001472.SED'], version='1.0', url=s11_website,
    type='SN IIP', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2005hl', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-002000.SED'], version='1.0', url=s11_website,
    type='SN Ib', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2005hm', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-002744.SED'], version='1.0', url=s11_website,
    type='SN Ib', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2005gi', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-003818.SED'], version='1.0', url=s11_website,
    type='SN IIP', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2006fo', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-013195.SED'], version='1.0', url=s11_website,
    type='SN Ic', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2006jo', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-014492.SED'], version='1.0', url=s11_website,
    type='SN Ib', subclass=s11_subclass, reference=s11_ref, note=s11_note)

registry.register_loader(
    Model, 's11-2006jl', load_timeseries_ascii_sncc,
    [s11_baseurl + 'S11_SDSS-014599.SED'], version='1.0', url=s11_website,
    type='SN IIP', subclass=s11_subclass, reference=s11_ref, note=s11_note)

del s11_baseurl
del s11_ref
del s11_website
del s11_subclass
del s11_note

# -----------------------------------------------------------------------
# Hsiao models

def load_stretchmodel_fits(remote_url, name=None, version=None):
    fn = download_file(remote_url, cache=True)
    hdulist = fits.open(fn)
    w = wcs.WCS(hdulist[0].header)
    flux_density = hdulist[0].data
    hdulist.close()
    
    ny, nx = flux_density.shape
    xcoords = np.arange(nx)
    ycoords = np.arange(ny)

    coords = np.swapaxes([xcoords, np.zeros(nx)], 0, 1)
    dispersion = w.wcs_pix2world(coords, 0)[:,0]

    coords = np.swapaxes([np.zeros(ny), ycoords], 0, 1)
    phase = w.wcs_pix2world(coords, 0)[:,1]
    model = StretchModel(phase, dispersion, flux_density, name=name,
                         version=version)
    set_bandfluxerror_sn1a(model)
    return model

hsiao_baseurl = 'http://kbarbary.github.com/data/models/'
hsiao_website = 'http://csp.obs.carnegiescience.edu/data/snpy'
hsiao_subclass = '`~sncosmo.StretchModel`'
hsiao_ref = ('H07', 'Hsiao et al. 2007 <http://adsabs.harvard.edu/abs/'
             '2007ApJ...663.1187H>')
hsiao_note = 'extracted from the SNooPy package on 21 Dec 2012.'

registry.register_loader(
    Model, 'hsiao', load_stretchmodel_fits, [hsiao_baseurl+'Hsiao_SED.fits'],
    version='1.0', url=hsiao_website, type='SN Ia', subclass=hsiao_subclass,
    reference=hsiao_ref, note=hsiao_note)
registry.register_loader(
    Model, 'hsiao', load_stretchmodel_fits,[hsiao_baseurl+'Hsiao_SED_V2.fits'],
    version='2.0', url=hsiao_website, type='SN Ia', subclass=hsiao_subclass,
    reference=hsiao_ref, note=hsiao_note)
registry.register_loader(
    Model, 'hsiao', load_stretchmodel_fits,[hsiao_baseurl+'Hsiao_SED_V3.fits'],
    version='3.0', url=hsiao_website, type='SN Ia', subclass=hsiao_subclass,
    reference=hsiao_ref, note=hsiao_note)

del hsiao_baseurl
del hsiao_website
del hsiao_subclass
del hsiao_ref
del hsiao_note

# -----------------------------------------------------------------------
# SALT2 models

def load_salt2model(remote_url, topdir, name=None, version=None):
    fn = download_file(remote_url, cache=True)
    t = tarfile.open(fn, 'r:gz')

    errscalefn = join(topdir, 'salt2_spec_dispersion_scaling.dat')
    if errscalefn in t.getnames():
        errscalefile = t.extractfile(errscalefn)
    else:
        errscalefile = None

    model = SALT2Model(
        m0file=t.extractfile(join(topdir,'salt2_template_0.dat')),
        m1file=t.extractfile(join(topdir,'salt2_template_1.dat')),
        v00file=t.extractfile(join(topdir,'salt2_spec_variance_0.dat')),
        v11file=t.extractfile(join(topdir,'salt2_spec_variance_1.dat')),
        v01file=t.extractfile(join(topdir,'salt2_spec_covariance_01.dat')),
        errscalefile=errscalefile, name=name, version=version)
    t.close()
    set_bandfluxerror_sn1a(model)
    return model

salt2_baseurl = 'http://supernovae.in2p3.fr/~guy/salt/download/'
salt2_website = 'http://supernovae.in2p3.fr/~guy/salt/download_templates.html'
salt2_reference = ('G07', 'Guy et al. 2007 '
                   '<http://adsabs.harvard.edu/abs/2007A%26A...466...11G>')
registry.register_loader(
    Model, 'salt2', load_salt2model,
    [salt2_baseurl + 'salt2_model_data-1-1.tar.gz', 'salt2-1-1'],
    version='1.1', type='SN Ia', subclass='`~sncosmo.SALT2Model`', 
    url=salt2_website, reference=salt2_reference)
registry.register_loader(
    Model, 'salt2', load_salt2model,
    [salt2_baseurl + 'salt2_model_data-2-0.tar.gz', 'salt2-2-0'],
    version='2.0', type='SN Ia', subclass='`~sncosmo.SALT2Model`', 
    url=salt2_website, reference=salt2_reference)


# --------------------------------------------------------------------------
# Generate docstring

lines = [
    '',
    '  '.join([16*'=', 7*'=', 8*'=', 27*'=', 14*'=', 7*'=', 7*'=']),
    '{:16}  {:7}  {:8}  {:27}  {:14}  {:7}  {:50}'.format(
        'Name', 'Version', 'Type', 'Subclass', 'Reference', 'Website', 'Notes')
    ]
lines.append(lines[1])

urlnums = {}
allnotes = []
allrefs = []
for m in registry.get_loaders_metadata(Model):

    reflink = ''
    urllink = ''
    notelink = ''

    if 'note' in m:
        if m['note'] not in allnotes: allnotes.append(m['note'])
        notenum = allnotes.index(m['note'])
        notelink = '[{}]_'.format(notenum + 1)

    if 'reference' in m:
        reflink = '[{}]_'.format(m['reference'][0])
        if m['reference'] not in allrefs:
            allrefs.append(m['reference'])

    if 'url' in m:
        url = m['url']
        if url not in urlnums:
            if len(urlnums) == 0: urlnums[url] = 0
            else: urlnums[url] = max(urlnums.values()) + 1
        urllink = '`{}`_'.format(string.letters[urlnums[url]])

    lines.append("{0!r:16}  {1!r:7}  {2:8}  {3:27}  {4:14}  {5:7}  {6:50}"
                 .format(m['name'], m['version'], m['type'], m['subclass'],
                         reflink, urllink, notelink))

lines.extend([lines[1], ''])
for refkey, ref in allrefs:
    lines.append('.. [{}] `{}`__'.format(refkey, ref))
lines.append('')
for url, urlnum in urlnums.iteritems():
    lines.append('.. _`{}`: {}'.format(string.letters[urlnum], url))
lines.append('')
for i, note in enumerate(allnotes):
    lines.append('.. [{}] {}'.format(i + 1, note))
lines.append('')
__doc__ = '\n'.join(lines)

del lines
del urlnums
del allrefs



# Autogenerated by Astropy-affiliated package sncosmo's setup.py on 2013-07-14 13:51:55.382862

from astropy.version_helpers import update_git_devstr, get_git_devstr

_last_generated_version = '0.1'

version = update_git_devstr(_last_generated_version)
githash = get_git_devstr(sha=True, show_warning=False)

major = 0
minor = 1
bugfix = 0

release = True
debug = False

try:
    from astropy.utils._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from astropy.cython_version import cython_version
except ImportError:
    cython_version = "unknown"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Importing this module registers loaders for built-in bandpasses.
# The module docstring, a table of the bandpasses, is generated at the
# after all the bandpasses are registered.

import string
from os.path import join

from astropy.io import ascii
from astropy import units as u
from astropy.utils.data import (download_file, get_pkg_data_filename,
                                get_readable_fileobj)
from astropy.utils import OrderedDict
from astropy.config import ConfigurationItem

from .. import registry
from .. import Bandpass

def load_bandpass_ascii(pkg_data_name, name=None):
    """Read two-column bandpass. First column is assumed to be wavelength
    in Angstroms."""
    
    filename = get_pkg_data_filename(pkg_data_name)
    t = ascii.read(filename, names=['disp', 'trans'])
    return Bandpass(t['disp'], t['trans'], dunit=u.AA, name=name)

# --------------------------------------------------------------------------
# DES

decam_url = 'http://www.ctio.noao.edu/noao/content/dark-energy-camera-decam'
decam_retrieved = '19 June 2012'
registry.register_loader(Bandpass, 'desg', load_bandpass_ascii,
                         ['../data/bandpasses/des_g.dat'], filterset='des',
                         description='Dark Energy Camera g band',
                         dataurl=decam_url, retrieved=decam_retrieved)
registry.register_loader(Bandpass, 'desr', load_bandpass_ascii,
                         ['../data/bandpasses/des_r.dat'], filterset='des',
                         description='Dark Energy Camera r band',
                         dataurl=decam_url, retrieved=decam_retrieved)
registry.register_loader(Bandpass, 'desi', load_bandpass_ascii,
                         ['../data/bandpasses/des_i.dat'], filterset='des',
                         description='Dark Energy Camera i band',
                         dataurl=decam_url, retrieved=decam_retrieved)
registry.register_loader(Bandpass, 'desz', load_bandpass_ascii,
                         ['../data/bandpasses/des_z.dat'], filterset='des',
                         description='Dark Energy Camera z band',
                         dataurl=decam_url, retrieved=decam_retrieved)
registry.register_loader(Bandpass, 'desy', load_bandpass_ascii,
                         ['../data/bandpasses/des_y.dat'], filterset='des',
                         description='Dark Energy Camera y band',
                         dataurl=decam_url, retrieved=decam_retrieved)
del decam_url
del decam_retrieved

# --------------------------------------------------------------------------
# Bessel 1990

bessell_ref = ('B90',
               '`Bessell 1990 <http://adsabs.harvard.edu/'
               'abs/1990PASP..102.1181B>`__, Table 2')
bessell_desc = 'Representation of Johnson-Cousins UBVRI system'

registry.register_loader(
    Bandpass, 'bessellux', load_bandpass_ascii,
    ['../data/bandpasses/bessell_ux.dat'], filterset='bessell',
    description=bessell_desc, reference=bessell_ref)
registry.register_loader(
    Bandpass, 'bessellb', load_bandpass_ascii,
    ['../data/bandpasses/bessell_b.dat'], filterset='bessell',
    description=bessell_desc, reference=bessell_ref)
registry.register_loader(
    Bandpass, 'bessellv', load_bandpass_ascii,
    ['../data/bandpasses/bessell_v.dat'], filterset='bessell',
    description=bessell_desc, reference=bessell_ref)
registry.register_loader(
    Bandpass, 'bessellr', load_bandpass_ascii,
    ['../data/bandpasses/bessell_r.dat'], filterset='bessell',
    description=bessell_desc, reference=bessell_ref)
registry.register_loader(
    Bandpass, 'besselli', load_bandpass_ascii,
    ['../data/bandpasses/bessell_i.dat'], filterset='bessell',
    description=bessell_desc, reference=bessell_ref)

del bessell_ref
del bessell_desc

# --------------------------------------------------------------------------
# SDSS

sdss_ref = ('D10', 
            '`Doi et al. 2010 '
            '<http://adsabs.harvard.edu/abs/2010AJ....139.1628D>`__, Table 4')
sdss_desc = \
    'SDSS 2.5m imager at airmass 1.3 (including atmosphere), normalized.'

registry.register_loader(
    Bandpass, 'sdssu', load_bandpass_ascii,
    ['../data/bandpasses/sdss_u.dat'], filterset='sdss',
    description=sdss_desc, reference=sdss_ref)
registry.register_loader(
    Bandpass, 'sdssg', load_bandpass_ascii,
    ['../data/bandpasses/sdss_g.dat'], filterset='sdss',
    description=sdss_desc, reference=sdss_ref)
registry.register_loader(
    Bandpass, 'sdssr', load_bandpass_ascii,
    ['../data/bandpasses/sdss_r.dat'], filterset='sdss',
    description=sdss_desc, reference=sdss_ref)
registry.register_loader(
    Bandpass, 'sdssi', load_bandpass_ascii,
    ['../data/bandpasses/sdss_i.dat'], filterset='sdss',
    description=sdss_desc, reference=sdss_ref)
registry.register_loader(
    Bandpass, 'sdssz', load_bandpass_ascii,
    ['../data/bandpasses/sdss_z.dat'], filterset='sdss',
    description=sdss_desc, reference=sdss_ref)

del sdss_ref
del sdss_desc

# --------------------------------------------------------------------------
# Generate docstring

lines = [
    '',
    '  '.join([11*'=', 80*'=', 14*'=', 8*'=', 12*'=']),
    '{:11}  {:80}  {:14}  {:8}  {:12}'
    .format('Name', 'Description', 'Reference', 'Data URL', 'Retrieved')
    ]
lines.append(lines[1])

urlnums = {}
allrefs = []
for m in registry.get_loaders_metadata(Bandpass):

    reflink = ''
    urllink = ''
    retrieved = ''

    if 'reference' in m:
        reflink = '[{}]_'.format(m['reference'][0])
        if m['reference'] not in allrefs:
            allrefs.append(m['reference'])

    if 'dataurl' in m:
        dataurl = m['dataurl']
        if dataurl not in urlnums:
            if len(urlnums) == 0: urlnums[dataurl] = 0
            else: urlnums[dataurl] = max(urlnums.values()) + 1
        urllink = '`{}`_'.format(string.letters[urlnums[dataurl]])

    if 'retrieved' in m:
        retrieved = m['retrieved']

    lines.append("{0!r:11}  {1:80}  {2:14}  {3:8}  {4:12}".format(
            m['name'], m['description'], reflink, urllink, retrieved))

lines.extend([lines[1], ''])
for refkey, ref in allrefs:
    lines.append('.. [{}] {}'.format(refkey, ref))
lines.append('')
for url, urlnum in urlnums.iteritems():
    lines.append('.. _`{}`: {}'.format(string.letters[urlnum], url))
lines.append('')
__doc__ = '\n'.join(lines)

del lines
del urlnums
del allrefs



from sys import stdout
import time
import math
import copy
import operator

import numpy as np
from scipy import integrate, optimize
from astropy.utils import OrderedDict

from .models import get_model
from .spectral import get_magsystem
from . import nest

def _cdf(pdf, x, a):
    return integrate.quad(pdf, a, x)[0]

def _ppf_to_solve(x, pdf, q, a):
    return _cdf(pdf, x, a) - q

def _ppf_single_call(pdf, q, a, b):
    left = right = None
    if a > -np.inf: left = a
    if b < np.inf: right = b

    factor = 10.

    # if lower limit is -infinity, adjust to
    # ensure that cdf(left) < q
    if  left is None:
        left = -1. * factor
        while _cdf(pdf, left, a) > q:
            right = left
            left *= factor

    # if upper limit is infinity, adjust to
    # ensure that cdf(right) > q
    if  right is None:
        right = factor
        while _cdf(pdf, right, a) < q:
            left = right
            right *= factor

    return optimize.brentq(_ppf_to_solve, left, right, args=(pdf, q, a))

class Interp1d(object):
    def __init__(self, xmin, xmax, y):
        self._xmin = xmin
        self._xmax = xmax
        self._n = len(y)
        self._xstep = (xmax - xmin) / (self._n - 1)
        self._y = y

    def __call__(self, x):
        """works only in range [xmin, xmax)"""
        nsteps = (x - self._xmin) / self._xstep
        i = int(nsteps)
        w = nsteps - i
        return (1.-w) * self._y[i] + w * self._y[i+1]

def pdf_to_ppf(pdf, a, b):
    """Given a function representing a pdf, return a callable representing the
    inverse cdf (or ppf) of the pdf."""

    n = 101
    x = np.linspace(0., 1., n)
    y = np.empty(n, dtype=np.float)
    y[0] = a
    y[-1] = b
    for i in range(1, n-1):
        y[i] = _ppf_single_call(pdf, x[i], a, b)

    return Interp1d(0., 1., y)


def evidence(model, data, parnames,
             parlims=None, priors=None, ppfs=None, tied=None,
             include_error=False, nobj=50, maxiter=10000,
             verbose=False, verbose_name=''):

    # Construct a list of ppfs to be used in the prior() function...
    npar = len(parnames)
    ppflist = npar * [None]

    # ... if a ppf is directly supplied for a parameter, it takes precedence...
    if ppfs is not None:
        for i, parname in enumerate(parnames):
            if parname in ppfs:
                ppflist[i] = ppfs[parname]

    # ...and for the parameters without ppfs, construct one from limits/prior.
    for i, parname in enumerate(parnames):
        if ppflist[i] is not None:
            continue
        if parname not in parlims:
            raise ValueError("Must supply ppf or limits for parameter '{}'"
                             .format(parname))
        a, b = parlims[parname]
        if (priors is not None and parname in priors):
            ppflist[i] = pdf_to_ppf(priors[parname], a, b)
        else:
            ppflist[i] = Interp1d(0., 1., np.array([a, b]))

    def prior(u):
        v = np.empty(npar, dtype=np.float)
        for i in range(npar):
            v[i] = ppflist[i](u[i])
        return v

    band = data['band']
    time = data['time']
    zp = data['zp']
    zpsys = data['zpsys']
    flux = data['flux']
    fluxerr = data['fluxerr']

    def loglikelihood(parvals):
        d = dict(zip(parnames, parvals))
        if tied is not None:
            for parname, func in tied.iteritems():
                d[parname] = func(d)

        model.set(**d)
        if not include_error:
            modelflux = model.bandflux(band, time, zp, zpsys)
            chisq = np.sum(((flux - modelflux) / fluxerr)**2)
        else:
            modelflux, modelfluxerr =  model.bandflux(band, time, zp, zpsys,
                                                      include_error=True)
            chisq = np.sum(((flux - modelflux)**2 /
                            (fluxerr**2 + modelfluxerr**2)))

        return -chisq / 2.

    res = nest.nest(loglikelihood, prior, npar, nobj=nobj, maxiter=maxiter,
                    verbose=verbose, verbose_name=verbose_name)

    # Add tied parameters to results. This is inelegant, but, eh.
    nsamples = len(res['samples_parvals'])
    res['nsamples'] = nsamples
    if tied is not None:
        tiedparnames = tied.keys()
        ntiedpar = len(tiedparnames)
        tiedparvals = np.empty((nsamples, ntiedpar), dtype=np.float)
        for i in range(nsamples):
            d = dict(zip(parnames, res['samples_parvals'][i, :]))
            for j, parname in enumerate(tiedparnames):
                tiedparvals[i, j] = tied[parname](d)

        res['samples_parvals'] = np.hstack((res['samples_parvals'], 
                                            tiedparvals))
        parnames = parnames + tiedparnames

    # Sample averages and their standard deviations.
    res['parvals'] = np.average(res['samples_parvals'],
                                weights=res['samples_wt'], axis=0)
    res['parerrs'] = np.sqrt(np.sum(res['samples_wt'][:, np.newaxis] *
                             res['samples_parvals']**2, axis=0) -
                             res['parvals']**2)

    # Add some more to results
    res['parnames'] = parnames
    res['chisq_min'] = -2. * res.pop('loglmax')
    res['dof'] = len(time) - npar

    return res


class PhotoTyper(object):
    """Baysian photometric typer.
    
    Parameters
    ----------
    verbose : bool, optional
        Print lines as evidence is calculated.
    t0range : tuple of floats: (t0_low, t0_high), optional
        Lower limit on t0 relative to earliest data point, and upper limit
        on t0 relative to latest data point, in days. Default is (0., 0.).

    Notes
    -----
    Parameters can also be set after initialization with, e.g.,
    ``typer.verbose = True`` or ``typer.t0range = (-10., 10.)``.
    """

    def __init__(self, verbose=True, t0range=(0., 0.)):
        self._models = OrderedDict()
        self.types = []
        self.verbose = verbose
        self.t0range = t0range

    def add_model(self, model, model_type, parlims, priors=None,
                  model_prior=1., tied=None, include_error=False,
                  name=None):
        """Add a model.

        Parameters
        ----------
        model : `~sncosmo.Model` or str
            A model instance or name of a model in the registry. 
        model_type : str
            A string identifier for the type of the model (e.g., 'SN Ia',
            'SN IIP', etc). Models of the same type are included together
            in determining probabilities.
        parlims : dict, optional
            Dictionary.
        tied : dict, optional
            Dictionary of functions, default is `None`.
        name : str
            Name 
        """

        model = get_model(model)  # This always copies the model.
        if name is None:
            name = model.name
        if name is None:
            raise ValueError('Model does not have a name. Set the name.')
        if name in self._models:
            raise ValueError('model of this name already included.')
        if model_prior <= 0.:
            raise ValueError('model_prior must be positive')

        # get ppf for each parameter
        ppfs = {}
        for parname, parlim in parlims.iteritems():
            a, b = parlims[parname]
            if (priors is not None and parname in priors):
                ppfs[parname] = pdf_to_ppf(priors[parname], a, b)
            else:
                ppfs[parname] = Interp1d(0., 1., np.array([a, b]))

        # Add model info to internal data.
        self._models[name] = {
            'model': model,
            'type': model_type,
            'parlims': parlims,
            'ppfs': ppfs,
            'tied': tied,
            'model_prior': model_prior,
            'include_error': include_error
            }

        # Add model type to list of types.
        if model_type not in self.types:
            self.types.append(model_type)

    def __str__(self):
        lines = []
        for model_type in self.types:
            lines.append('Type: {}'.format(model_type))
            for name, d in self._models.iteritems():
                if d['type'] != model_type: continue
                lines.append('  Model: {}'.format(name))
                for parname, parvals in d['parlims'].iteritems():
                    lines.append('    {}: [{} .. {}]'
                                 .format(parname, parvals[0], parvals[1]))
        return '\n'.join(lines)

    def classify(self, data, verbose=None):
        """Determine probability of each model type for the given data.

        Parameters
        ----------
        data : `~numpy.ndarray` or dict
            Light curve data to classify.
        verbose : bool, optional
            If True, print information during iteration. (If False, don't).
            Default is to use the value of self.verbose.

        Returns
        -------
        types : OrderedDict
            Keys are type names ('SN Ia'), values are dictionaries
            containing key 'p'. Entries are sorted by 'p', decreasing.
        models : OrderedDict
            Keys are model names. Values are dictionaries with the
            following keys:
           
            * 'p' : probability (float)
            * 'perr' : numerical error on p (tuple)
            * 'type' : Type of model (str)
            * 'dof' : degrees of freedom (len(data) - npar)
            * 'niter' : number of iterations (int)
            * 'ncalls' : number of likelihood calls (int)
            * 'time' : evaluation time in seconds (float)
            * 'parnames': parameter names (list of str)
            * 'parvals': parameter "best-fit" values (list of float)
            * 'parerrs': parameter "errors" (list of float)
            * 'chisq_min': minimum chi^2 value of any sample (float)
            * 'samples_parvals' : ndarray, shape=(nsamples, npars)
            * 'samples_wt': ndarray, shape=(nsamples,)

            Entries are sorted by 'p', decreasing.
        """

        if verbose is None:
            verbose = self.verbose

        # limit data to bands that overlap *all* models over the full z range.
        valid = np.ones(len(data['band']), dtype=np.bool)
        for m in self._models.values():
            model = m['model']
            if 'z' not in m['parlims']:
                v = model.bandoverlap(data['band'])
            else:
                v = np.all(model.bandoverlap(data['band'],z=m['parlims']['z']),
                           axis=1)
            valid = valid & v
        if not np.all(valid):
            print "WARNING: dropping following bands from data:"
            print np.unique(data['band'][np.invert(valid)])
            data = {'time': data['time'][valid],
                    'band': data['band'][valid],
                    'flux': data['flux'][valid],
                    'fluxerr': data['fluxerr'][valid],
                    'zp': data['zp'][valid],
                    'zpsys': data['zpsys'][valid]}

        # get range of t0 to consider
        parlims = {
            't0': (np.min(data['time']) + self.t0range[0],
                   np.max(data['time']) + self.t0range[1])
            }

        models = {}
        for name, m in self._models.iteritems():
            parnames = m['ppfs'].keys() + ['t0']
            models[name] = evidence(
                m['model'], data, parnames,
                parlims=parlims, ppfs=m['ppfs'], tied=m['tied'],
                include_error=m['include_error'],
                verbose=verbose, verbose_name=name)

            # multiply evidence by model prior
            models[name]['logz'] += m['model_prior']

            # add type info
            models[name]['type'] = m['type']

        # get denominator (sum of Z)
        logzvals = [d['logz'] for d in models.values()]
        logzsum = logzvals[0]
        for i in range(1, len(logzvals)):
            logzsum = np.logaddexp(logzsum, logzvals[i])
        
        # get probability of each model: p = Z_i / sum(Z)
        # Errors are calculated by finding the upper and lower Z limits:
        # up   = exp(logz + logzerr) = exp(logz)exp(logzerr) = z*exp(logzerr)
        # down = exp(logz - logzerr) = exp(logz)/exp(logzerr) = z/exp(logzerr)
        for name, d in models.iteritems():
            p = np.exp(d['logz'] - logzsum)
            explogzerr = math.exp(d['logzerr'])
            p_up = p * explogzerr / (1. - p + p * explogzerr)
            p_dn = p / explogzerr / (1. - p + p / explogzerr)
            d['p'] = p
            d['perr'] = (p_up - p, p - p_dn)

        # get probability for each type
        types = {name: {'p': 0.} for name in self.types}
        for name, d in models.iteritems():
            types[d['type']]['p'] += d['p']

        # sort models and types
        types = OrderedDict(
            sorted(types.items(), key=lambda t: t[1]['p'], reverse=True)
            )
        models = OrderedDict(
            sorted(models.items(), key=lambda t: t[1]['p'], reverse=True)
            )

        return types, models



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Classes to represent spectral models of astronomical transients."""

import abc
import os
import math
import copy
from itertools import product
from textwrap import dedent

import numpy as np
from scipy.interpolate import RectBivariateSpline as Spline2d
from scipy.interpolate import InterpolatedUnivariateSpline as Spline1d
from scipy.ndimage.interpolation import map_coordinates as mapcoords

import astropy.constants as const
import astropy.units as u
from astropy.utils import OrderedDict
from astropy.utils.misc import isiterable
from astropy import cosmology

from .io import read_griddata
from .extinction import extinction_ccm
from .spectral import Spectrum, Bandpass, MagSystem, get_bandpass, get_magsystem
from . import registry

#import models_interp

__all__ = ['get_model', 'Model', 'TimeSeriesModel', 'StretchModel',
           'SALT2Model']

cosmology.set_current(cosmology.WMAP9)
HC_ERG_AA = const.h.cgs.value * const.c.to(u.AA / u.s).value

def get_model(name, version=None):
    """Retrieve a model from the registry by name.

    Parameters
    ----------
    name : str
        Name of model in the registry.
    version : str, optional
        Version identifier for models with multiple versions. Default is
        `None` which corresponds to the latest, or only, version.
    """
    
    # Call the one in the registry to create a copy and keep the registry
    # copy pristine.
    return registry.retrieve(Model, name, version=version)()


class Model(object):
    """An abstract base class for transient models.
    
    A "transient model" in this case is the spectral time evolution
    of a source as a function of an arbitrary number of parameters.

    This is an abstract base class -- You can't create instances of
    this class. Instead, you must work with subclasses such as
    `TimeSeriesModel`. Subclasses must define (at minimum) `__init__()` and
    the private method `_model_flux()` 
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self, name=None, version=None):
        self._params = OrderedDict(
            [('fscale', 1.), ('m', None), ('mabs', None), ('t0', 0.),
             ('z', None)])
        self._refphase = 0.
        self._refband = get_bandpass('bessellb')
        self._refmagsys = get_magsystem('ab')
        self._refm = None
        self._cosmo = cosmology.get_current()
        self._distmod = None  # Distance modulus
        self.name = name
        self.version = version

        self._model_bandfluxrelerr = None

    def set(self, **params):
        """Set the parameters of the model using keyword values.

        Parameters
        ----------
        fscale : float
            The internal model flux spectral density values will be multiplied
            by this factor. If specified, `mabs` and `m` become unknown and
            are set to `None`.
        m : float
            Apparent magnitude; alternative way to set `fscale` (overrides it).
        mabs : float
            Absolute magnitude; alternative way to set `m` (overrides it).
            Apparent magnitude `m` is set according to the luminosity distance.
            Only permitted if luminosity distance is known (if cosmology is
            not None and z is not None).
        t0 : float
            Time offset.
        z : float
            Redshift.

        Notes
        -----
        `fscale` vs `mabs` vs `m` are various ways to set `fscale`;
        only one may be used at a time.
        """

        # Below, when `self._refm` and `self._distmod` might have changed, we
        # set them to `None` so that they are recalculated as needed.

        if 'z' in params and self._params['z'] != params['z']:
            self._distmod = None

        # set parameters and note whether there are any that might change
        # spectral shape.
        for key, val in params.items():
            if key not in self._params:
                continue
            if key != 't0':
                recalc_fscale = True
                if key not in ['z', 'fscale', 'm', 'mabs']:
                    self._refm = None
            self._params[key] = val

        # In what we just set, 'mabs' overrides 'm' overrides 'fscale'.
        # Setting 'm' makes 'mabs' undefined, and setting 'fscale' makes
        # both 'm' and 'mabs' undefined.
        if 'mabs' in params:
            pass
        elif 'm' in params:
            self._params['mabs'] = None
        elif 'fscale' in params:
            self._params['m'] = None
            self._params['mabs'] = None

        if 'z' in params:
            self._distmod = None

        # set fscale from mabs or m
        if self._params['mabs'] is not None:
            if recalc_fscale:
                self._set_fscale_from_mabs()
        elif self._params['m'] is not None:
            if recalc_fscale:
                self._set_fscale_from_m()

    def _set_fscale_from_m(self):
        """Determine the fscale that when applied to the model
        (with current parameters), will result in the desired absolute
        magnitude at the reference phase."""

        if self._refm is None:
            fscale = self._params['fscale']
            self._params['fscale'] = 1.
            self._refm = self.bandmag(self._refband, self._refmagsys,
                                      self._refphase, modelframe=True)
            self._params['fscale'] = fscale

        self._params['fscale'] = 10.**(0.4 * (self._refm - self._params['m']))

    def _set_fscale_from_mabs(self):
        """Determine the fscale that when applied to the model
        (with current parameters), will result in the desired absolute
        magnitude at the reference phase."""

        # We need the distance modulus. Try to calculate it.
        if self._distmod is None:
            if self._params['z'] is None or self._cosmo is None:
                raise ValueError("cannot set absolute magnitude when distance "
                                 "modulus is unknown (when either redshift or "
                                 "cosmology is None)")
            self._distmod = self._cosmo.distmod(self._params['z'])

        self._params['m'] = self._params['mabs'] + self._distmod
        self._set_fscale_from_m()

    @abc.abstractmethod
    def _model_flux(self, phase, disp):
        """Return the model spectral flux density (without any scaling or
        redshifting.

        Parameters
        ----------
        phase : `~numpy.ndarray`
            1-d array of model phases in days. Must be within bounds of
            the model, and monotonically increasing.
        disp : `~numpy.ndarray`
            1-d array of wavelength values in Angstroms. Must be within
            bounds of the model and monotonically increasing.

        Returns
        -------
        flux : `~numpy.ndarray`
            2-d array of shape ``(len(phase), len(disp))`` giving model
            flux spectral density.
        """
        pass

    @property
    def parnames(self):
        return self._params.keys()

    @property
    def params(self):
        """Dictionary of model parameters. (Read-only.) Use `set` method to
        set parameters."""
        return copy.copy(self._params)

    @property
    def cosmo(self):
        """Cosmology used to calculate luminosity distance."""
        return self._cosmo

    @cosmo.setter
    def cosmo(self, new_cosmology):
        if (new_cosmology is not None and 
            not isinstance(new_cosmology, cosmology.FLRW)):
            raise ValueError('cosmology must be None or '
                             'astropy.cosmology.FLRW instance.')
        self._cosmo = new_cosmology
        self._distmod = None  # becomes unknown

    @property
    def refphase(self):
        """Phase at which absolute magnitude is evaluated."""
        return self._refphase

    @refphase.setter
    def refphase(self, new_refphase):
        self._refphase = new_refphase

    @property
    def refband(self):
        """Bandpass in absolute magnitude is evaluated."""
        return self._refband

    @refband.setter
    def refband(self, band):
        self._refband = get_bandpass(band)

    @property
    def refmagsys(self):
        """Magnitude system in which absolute magnitude is evaluated."""
        return self._refmagsys

    @refmagsys.setter
    def refmagsys(self, magsystem):
        self._refmagsys = get_magsystem(magsystem)


    def times(self, modelframe=False):
        """Return native time sampling of the model.

        Parameters
        ----------
        modelframe : bool, optional
            If True, return rest-frame phases. Default is False.
        """
        

        if modelframe:
            return self._phase
        
        z = self._params['z']
        if z is None: z = 0.

        return self._params['t0'] + (1. + z) * self._phase

    def disp(self, modelframe=False):
        """Return native dispersion grid of the model.

        Parameters
        ----------
        modelframe : bool, optional
            If True, return rest-frame phases. Default is False, which
            corresponds to observer-frame phases (model phases
            multiplied by 1 + z).
        """
        if modelframe or self._params['z'] is None:
            return self._disp
        else:
            return self._disp * (1. + self._params['z'])

    def flux(self, time=None, disp=None, modelframe=False):
        """The model flux spectral density at the given dispersion values.

        Parameters
        ----------
        time : float or list_like, optional
            Time(s) in days. Times are observer frame unless `modelframe`
            is True, in which case times are assumed to be model phases.
            If `None` (default) the native phases of the model are used.
        disp : float or list_like, optional
            Model dispersion values in Angstroms. Interpreted to be in the
            observer frame unless `modelframe` is True. If `None` (default)
            the native dispersion of the model is used.
        modelframe : bool, optional
            If True, return fluxes without redshifting or time shifting.

        Returns
        -------
        flux : float or `~numpy.ndarray`
            Spectral flux density values in ergs / s / cm^2 / Angstrom.
        """

        z = max(0., self._params['z'])

        if time is None:
            phase = self._phase
        else:
            phase = self._time_to_phase(np.asarray(time), modelframe)

        if disp is None:
            disp = self._disp
        else:
            disp = np.asarray(disp)
            if not modelframe:
                disp /= (1. + z)
            if np.any(disp < self._disp[0]) or np.any(disp > self._disp[-1]):
                raise ValueError('requested dispersion value(s) outside '
                                 'model range')

        # Determine flux scaling factor
        factor = self._params['fscale']
        if not modelframe:
            factor /= 1. + z

        # Check dimensions of phase, disp for return value
        # (1, 1) -> ndim=2
        # (1, 0) -> ndim=2
        # (0, 1) -> ndim=1
        # (0, 0) -> float
        flux = factor * self._model_flux(phase, disp)
        if phase.ndim == 0:
            if disp.ndim == 0:
                return flux[0, 0]
            return flux[0, :]
        return flux

    def bandoverlap(self, band, z=None):
        """Return True if model dispersion range fully overlaps the band.

        Parameters
        ----------
        band : `~sncosmo.Bandpass`, str or list_like
            Bandpass, name of bandpass in registry, or list or array thereof.
        z : float or list_like, optional
            If given, evaluate the overlap when the model is at the given
            redshifts. If `None`, use the model redshift.

        Returns
        -------
        overlap : bool or `~numpy.ndarray`
            
        """
        
        band = np.asarray(band)
        if z is None: z = self._params['z']
        if z is None: z = 0
        z = np.asarray(z)
        ndim = (band.ndim, z.ndim)
        band = band.ravel()
        z = z.ravel()
        overlap = np.empty((len(band), len(z)), dtype=np.bool)
        for i, b in enumerate(band):
            b = get_bandpass(b)
            overlap[i, :] = ((b.disp[0] > self._disp[0] * (1. + z)) &
                             (b.disp[-1] < self._disp[-1] * (1. + z)))
        if ndim == (0, 0):
            return overlap[0, 0]
        if ndim[1] == 0:
            return overlap[:, 0]
        return overlap

    def bandflux(self, band, time=None, zp=None, zpsys=None,
                 modelframe=False, include_error=False):
        """Flux through the given bandpass(es) at the given time(s).

        Default return value is flux in photons / s / cm^2. If zp and zpsys
        are given, flux(es) are scaled to the requested zeropoints.

        Parameters
        ----------
        band : `~sncosmo.Bandpass` or str or list_like
            Bandpass(es) or name(s) of bandpass(es) in registry.
        time : float or list_like, optional
            Time(s) in days. Default is `None`, which corresponds to the full
            native time sampling of the model.
        zp : float or list_like, optional
            If given, zeropoint to scale flux to. If `None` (default) flux
            is not scaled.
        zpsys : `~sncosmo.MagSystem` or str (or list_like), optional
            Determines the magnitude system of the requested zeropoint.
            Cannot be `None` if `zp` is not `None`.
        include_error : bool, optional
            Include flux errors in return value. Default is False.


        Returns
        -------
        bandflux : float or `~numpy.ndarray`
            Flux in photons / s /cm^2, unless `zp` and `zpsys` are
            given, in which case flux is scaled so that it corresponds
            to the requested zeropoint. Return value is `float` if all
            input parameters are scalars, `~numpy.ndarray` otherwise.
        bandfluxerr : float or `~numpy.ndarray`
        """

        z = max(0., self._params['z'])
        if modelframe: z = 0.

        # Convert times to model phases
        if time is None:
            phase = self._phase
        else:
            phase = self._time_to_phase(np.asarray(time), modelframe)

        # Determine flux scaling factor
        factor = self._params['fscale']
        if not modelframe:
            factor /= 1. + z

        # broadcast arrays
        if zp is None:
            phase, band = np.broadcast_arrays(phase, band)
        else:
            if zpsys is None:
                raise ValueError('zpsys must be given if zp is not None')
            phase, band, zp, zpsys = \
                np.broadcast_arrays(phase, band, zp, zpsys)
            zp = np.atleast_1d(zp)
            zpsys = np.atleast_1d(zpsys)

        # convert to 1d arrays
        ndim = phase.ndim # save input ndim for return val
        phase = np.atleast_1d(phase)
        band = np.atleast_1d(band)

        # initialize output arrays
        bandflux = np.zeros(phase.shape, dtype=np.float)
        if include_error:
            relerr = np.zeros(phase.shape, dtype=np.float)

        # index of times that are in model range
        idx_validphase = (phase >= self._phase[0]) & (phase <= self._phase[-1])

        # loop over unique bands
        for b in set(band):
            idx = (band == b) & idx_validphase
            if not np.any(idx):
                continue
            b = get_bandpass(b)
            d = b.disp / (1. + z)  # bandpass dispersion in modelframe
            
            # make sure bandpass dispersion is in model range.
            if (d[0] < self._disp[0] or d[-1] > self._disp[-1]):
                raise ValueError(
                    'bandpass {0!r:s} [{1:.6g}, .., {2:.6g}] '
                    'outside model range [{3:.6g}, .., {4:.6g}]'
                    .format(b.name, b.disp[0], b.disp[-1], 
                            (1.+z) * self._disp[0], (1.+z) * self._disp[-1]))
            flux = self._model_flux(phase[idx], d)
            tmp = b.trans * b.disp * b.ddisp
            fluxsum = np.sum(flux * tmp, axis=1) / HC_ERG_AA

            if zp is not None:
                zpnorm = 10. ** (0.4*zp[idx])
                bandzpsys = zpsys[idx]
                for ms in set(bandzpsys):
                    idx2 = bandzpsys == ms
                    ms = get_magsystem(ms)
                    zpnorm[idx2] = zpnorm[idx2] / ms.zpbandflux(b)
                fluxsum *= zpnorm

            bandflux[idx] = fluxsum

            if include_error:
                deff = b.disp_eff
                if not modelframe: deff /= (1.+z)
                relerr[idx] = self._model_bandfluxrelerr(phase[idx], deff)[:, 0]

        # multiply all by overall factor determined above.
        bandflux *= factor

        # Return result.
        if include_error:
            if ndim == 0:
                return bandflux[0], bandflux[0] * relerr[0]
            return bandflux, bandflux * relerr
        if ndim == 0:
            return bandflux[0]
        return bandflux

    def bandmag(self, band, magsys, time=None, modelframe=False):
        """Magnitude at the given time(s) through the given 
        bandpass(es), and for the given magnitude system(s).

        Parameters
        ----------
        time : float or list_like
            Time(s) in days.
        band : `Bandpass` or str (or list_like)
            Bandpass or name of bandpass in registry.
        magsys : `MagSystem` or str (or list_like)
            MagSystem or name of MagSystem in registry.

        Returns
        -------
        mag : float or `~numpy.ndarray`
            Magnitude for each item in time, band, magsys.
            The return value is a float if all parameters are not interables.
            The return value is an `~numpy.ndarray` if any are interable.
        """

        bandflux = self.bandflux(band, time, modelframe=modelframe)
        band, magsys, bandflux = np.broadcast_arrays(band, magsys, bandflux)
        return_scalar = (band.ndim == 0)
        band = band.ravel()
        magsys = magsys.ravel()
        bandflux = bandflux.ravel()

        result = np.empty(bandflux.shape, dtype=np.float)
        for i, (b, ms, f) in enumerate(zip(band, magsys, bandflux)):
            ms = get_magsystem(ms)
            zpf = ms.zpbandflux(b)
            result[i] = -2.5 * np.log10(f / zpf)

        if return_scalar:
            return result[0]
        return result

    def set_bandflux_relative_error(self, phase, disp, relative_error):
        """Set the relative error to be applied to the bandflux.

        Parameters
        ----------
        phase : `~numpy.ndarray` (1d)
            Phases in days.
        disp : `~numpy.ndarray` (1d)
            Wavelength values.
        relative_error : `~numpy.ndarray` (2d)
            Fractional error on bandflux. Must have shape ``(len(phases),
            len(disp))``
        """
        self._model_bandfluxrelerr = \
            Spline2d(phase, disp, relative_error, kx=1, ky=1)

    def _time_to_phase(self, time, modelframe):
        """Convert time(s) (ndarray) to phase(s) (ndarray)."""

        if modelframe:
            return time
        return (time - self._params['t0']) / (1. + max(0., self._params['z']))


    def _phase_of_max_flux(self, band):
        """Determine phase of maximum flux in the given band, by fitting
        a parabola to phase of maximum flux and the surrounding two
        phases."""

        fluxes = self.bandflux(band)
        i = np.argmax(fluxes)
        x = self._phase[i-1: i+2]
        y = fluxes[i-1: i+2]
        A = np.hstack([x.reshape(3,1)**2, x.reshape(3,1), np.ones((3,1))])
        a, b, c = np.linalg.solve(A, y)
        return -b / (2 * a)


    def __call__(self, **params):
        """Return a shallow copy of the model with parameters set.
        See `set` method for details on parameters.
        
        Returns
        -------
        model : `sncosmo.Model`
            Model instance with parameters set as requested.

        Examples
        --------
        
        >>> model = sncosmo.get_model('salt2')
        >>> sn = model(c=0.1, x1=0.5, mabs=-19.3, z=0.68)
        >>> sn.bandmag('desg', 'ab', 0.)
        25.577096820065883
        >>> sn = model(c=0., x1=0., mabs=-19.3, z=0.5)
        >>> sn.bandmag('desg', 'ab', 0.)
        23.73028907970092
        """

        model = copy.copy(self)
        model._params = copy.copy(self._params)
        model.set(**params)
        return model

    def __repr__(self):
        name = ''
        version = ''
        if self.name is not None:
            name = ' {0!r:s}'.format(self.name)
        if self.version is not None:
            version = ' version={0!r:s}'.format(self.version)
        return "<{0:s}{1:s}{2:s} at 0x{3:x}>".format(
            self.__class__.__name__, name, version, id(self))

    def __str__(self):
        dmstr = '--'
        ldstr = '--'
        if self._distmod is not None:
            dmstr = '{:.6g}'.format(self._distmod)
            ld =  10.**((self._distmod - 25.) / 5.)
            ldstr = '{:.6g} Mpc'.format(ld)
        result = """\
        Model class: {}
        Model name: {}
        Model version: {}
        Model phases: [{:.6g}, .., {:.6g}] days ({:d} points)
        Model dispersion: [{:.6g}, .., {:.6g}] Angstroms ({:d} points) 
        Reference phase: {:.5f} days
        Cosmology: {}
        Current Parameters:
        """.format(
            self.__class__.__name__,
            self.name, self.version,
            self._phase[0], self._phase[-1], len(self._phase),
            self._disp[0], self._disp[-1],
            len(self._disp),
            self._refphase,
            self._cosmo)
        result = dedent(result)

        parameter_lines = []
        for key, val in self._params.iteritems():
            line = '    {} = {}'.format(key, val)
            if key in ['m', 'mabs']:
                line += ' [{}, {}]'.format(self._refband.name,
                                           self._refmagsys.name)
            elif key == 'z' and val is not None:
                line += (' [dist. mod. = {}, lum. dist. = {}]'
                         .format(dmstr, ldstr))
            parameter_lines.append(line)
        return result + '\n'.join(parameter_lines)


class TimeSeriesModel(Model):
    """A single-component spectral time series model.

    Parameters
    ----------
    phase : `~numpy.ndarray`
        Phases in days.
    disp : `~numpy.ndarray`
        Wavelengths in Angstroms.
    flux : `~numpy.ndarray`
        Model spectral flux density in erg / s / cm^2 / Angstrom.
        Must have shape ``(num_phases, num_disp)``.
    extinction_func : function, optional
        A function that accepts an array of wavelengths in Angstroms
        and returns an array representing the ratio of total extinction
        (in magnitudes) to the parameter ``c`` at each wavelength. Default
        is `sncosmo.extinction_ccm`.
    extinction_kwargs : dict
        A dictionary of keyword arguments to pass to `extinction_func`.
        Default is ``{'ebv': 1.}``.
    name : str, optional
        Name of the model. Default is `None`.
    version : str, optional
        Version of the model. Default is `None`.
    """

    def __init__(self, phase, disp, flux,
                 extinction_func=extinction_ccm,
                 extinction_kwargs=dict(ebv=1.),
                 name=None, version=None):

        super(TimeSeriesModel, self).__init__(name, version)
        self._params['c'] = None

        self._phase = phase
        self._disp = disp
        self._flux = Spline2d(phase, disp, flux, kx=2, ky=2)
        self._model_bandfluxrelerr = Spline2d(phase[[0, -1]], disp[[0, -1]],
                                              np.zeros((2,2), dtype=np.float),
                                              kx=1, ky=1)
            
        self.set_extinction_func(extinction_func, extinction_kwargs)

        self._refphase = self._phase_of_max_flux(self._refband)

    def _model_flux(self, phase, disp):
        """Return the model flux density (without any scaling or redshifting).
        """
        if self._params['c'] is None:
            return self._flux(phase, disp)
        
        dust_trans = self._dust_trans_base(disp) ** self._params['c']
        return dust_trans * self._flux(phase, disp)

    def set_extinction_func(self, func, extra_params):
        """Set the extinction ratio function to use in the model.

        Parameters
        ----------
        func : function
            A function that accepts an array of wavelengths in Angstroms
            and returns the ratio of total extinction
            (in magnitudes) to the parameter `c` at each wavelength.
        extra_params : dict
            A dictionary of keyword arguments to pass to the function.
            Default is `None`.
        """
        
        # set these parameters for info
        self._extinction_func = func
        self._extinction_kwargs = extra_params

        if extra_params is None:
            ext_ratio = func(self._disp)
        else:
            ext_ratio = func(self._disp, **extra_params)

        # calculate extinction base values, so that
        # self._dust_trans_base ** c gives the dust transmission.
        dust_trans_base =  10. ** (-0.4 * ext_ratio)
        self._dust_trans_base = Spline1d(self._disp, dust_trans_base, k=2)


class StretchModel(TimeSeriesModel):
    """A single-component spectral time series model, that "stretches".

    Parameters
    ----------
    phase : `~numpy.ndarray`
        Phases in days.
    disp : `~numpy.ndarray`
        Wavelengths in Angstroms.
    flux : `~numpy.ndarray`
        Model spectral flux density in erg / s / cm^2 / Angstrom.
        Must have shape `(num_phases, num_disp)`.

    Other Parameters
    ----------------
    flux_error : `~numpy.ndarray`, optional
        Model error on `flux`. Must have same shape as `flux`.
        Default is `None`.
    extinction_func : function, optional
        A function that accepts an array of wavelengths in Angstroms
        and returns an array representing the ratio of total extinction
        (in magnitudes) to the parameter `c` at each wavelength. Default
        is `sncosmo.extinction.extinction_ccm`
    extinction_kwargs : dict
        A dictionary of keyword arguments to pass to `extinction_func`.
        Default is `dict(ebv=1.)`.
    name : str, optional
        Name of the model. Default is `None`.
    version : str, optional
        Version of the model. Default is `None`.
    """
    def __init__(self, phase, disp, flux,
                 name=None, version=None,
                 extinction_func=extinction_ccm,
                 extinction_kwargs=dict(ebv=1.)):

        super(StretchModel, self).__init__(
            phase, disp, flux,
            name=name, version=version,
            extinction_func=extinction_func,
            extinction_kwargs=extinction_kwargs)
        
        self._params['s'] = 1.

    def _set_fscale_from_m(self):
        """Need to override this so that the refphase is applied correctly.
        refphase refers to the *unstretched* phase, but bandmag() with
        modelframe=True will treat the input as the *stretched* phase.
        if we don't do this, the phase of peak will shift with s, if
        the peak is not at phase = 0."""

        s = self._params['s']
        self._params['s'] = 1. # temporarily remove stretch
        super(StretchModel, self)._set_fscale_from_m()
        self._params['s'] = s # put it back.

    def times(self, modelframe=False):
        """Return native times of the model.

        Parameters
        ----------
        modelframe : bool, optional
            If True, return rest-frame phases. Default is False, which
            corresponds to observer-frame phases (model phases
            multiplied by 1 + z).
        """
        if modelframe:
            return self._params['s'] * self._phase
        z = max(0., self._params['z'])
        return self._params['t0'] + self._params['s'] * (1.+z) * self._phase

    def _time_to_phase(self, time, modelframe):
        """Convert time(s) (ndarray) to phase(s) (ndarray)."""

        if modelframe:
            return time / self._params['s']
        return ((time - self._params['t0']) /
                (self._params['s'] * (1. + max(0., self._params['z']))))


class SALT2Model(Model):
    """The SALT2 Type Ia supernova spectral timeseries model.

    Parameters
    ----------
    modeldir : str, optional
        Path to files containing model components. Each file should have
        the format:
 
            phase wavelength value

        on each line. If you want to give the absolute path to each file,
        set to `None`.
    m0file, m1file, v00file, v11file, v01file : str or fileobj, optional
        Filenames of various model components. Defaults are:

        * m0file = 'salt2_template_0.dat'
        * m1file = 'salt2_template_1.dat'
        * v00file = 'salt2_spec_variance_0.dat'
        * v11file = 'salt2_spec_variance_1.dat'
        * v01file = 'salt2_spec_covariance_01.dat'

    errscalefile : str, optional
        Name of error scale file, same format as model component files.
        The default is ``None``, which means that the error scale will
        not be applied in the ``fluxerr()`` method. This is only used for
        template versions 1.1 and 1.0, not 2.0+.

    Notes
    -----
    The phase and wavelength values of the various components don't necessarily
    need to match. (In the most recent salt2 model data, they do not all 
    match.) The phase and wavelength values of the first model component
    (in ``m0file``) are taken as the "native" sampling of the model, even
    though these values might require interpolation of the other model
    components.
    """

    def __init__(self, modeldir=None,
                 m0file='salt2_template_0.dat',
                 m1file='salt2_template_1.dat',
                 v00file='salt2_spec_variance_0.dat',
                 v11file='salt2_spec_variance_1.dat',
                 v01file='salt2_spec_covariance_01.dat',
                 errscalefile=None, name=None, version=None):
        super(SALT2Model, self).__init__(name, version)
        self._params['c'] = 0.
        self._params['x1'] = 0.
        self._model = {}

        components = ['M0', 'M1', 'V00', 'V11', 'V01', 'errscale']
        names_or_objs = [m0file, m1file, v00file, v11file, v01file,
                         errscalefile]

        # Make filenames into full paths.
        if modeldir is not None:
            for i in range(len(names_or_objs)):
                if (names_or_objs[i] is not None and
                    isinstance(names_or_objs[i], basestring)):
                    names_or_objs[i] = os.path.join(modeldir, names_or_objs[i])

        for component, name_or_obj in zip(components, names_or_objs):

            # If the filename is None, that component is left out of the model
            if name_or_obj is None: continue

            # Get the model component from the file
            phase, wavelength, values = read_griddata(name_or_obj)
            self._model[component] = Spline2d(phase, wavelength, values,
                                              kx=2, ky=2)

            # The "native" phases and wavelengths of the model are those
            # of the first model component.
            if component == 'M0':
                self._phase = phase
                self._disp = wavelength

        # add extinction component
        ext_ratio = self._extinction(self._disp)
        dust_trans_base = 10. ** (-0.4 * ext_ratio)
        self._model['ext'] = Spline1d(self._disp, dust_trans_base, k=2)

        # Set relative bandflux error
        self._model_bandfluxrelerr = \
            Spline2d(self._phase[[0, -1]], self._disp[[0, -1]], 
                     np.zeros((2,2), dtype=np.float), kx=1, ky=1)

        # set refphase
        self._refphase = self._phase_of_max_flux(self._refband)

    def _model_flux(self, phase, disp):
        """Return the model flux density (without any scaling or redshifting).
        """

        f0 = self._model['M0'](phase, disp)
        f1 = self._model['M1'](phase, disp)
        flux = f0 + self._params['x1'] * f1
        flux *= self._model['ext'](disp) ** self._params['c']
        return flux

    # No longer supporting spectral flux error, so we comment out the 
    # following method. This might be added back at a later date if
    # I gain a better understanding of how to use the SALT2 model errors.

    #def _model_fluxerr(self, phase=None, disp=None):
    #    """Return the model flux density error (without any scaling or
    #    redshifting. Return `None` if the error is undefined.
    #    """
    #    if phase is None:
    #        phase = self._phase
    #    if disp is None:
    #        disp = self._disp
    #    v00 = self._model['V00'](phase, disp)
    #    v11 = self._model['V11'](phase, disp)
    #    v01 = self._model['V01'](phase, disp)

        #TODO apply x0 correctly
    #    x1 = self._params['x1']
    #    sigma = np.sqrt(v00 + x1**2 * v11 + 2 * x1 * v01)
    #    sigma *= self._model['ext'](disp) ** self._params['c']
        ### sigma *= 1e-12   #- Magic constant from SALT2 code
        
    #    if 'errscale' in self._model:
    #        sigma *= self._model['errscale'](phase, disp)
        
        # Hack adjustment to error (from SALT2 code)
        #TODO: figure out a way to do this.
        #if phase < -7. or phase > 12.:
        #    idx = (disp < 3400.)
        #    sigma[idx] *= 1000.
        
    #   return sigma


    def _extinction(self, wavelengths, params=[-0.336646, 0.0484495]):
        """Return the extinction as a function of wavelength, for c=1.

        Notes
        -----
        From SALT2 code comments:

            ext = exp(color * constant * 
                      (l + params(0)*l^2 + params(1)*l^3 + ... ) /
                      (1 + params(0) + params(1) + ... ) )
                = exp(color * constant *  numerator / denominator )
                = exp(color * expo_term ) 
        """

        wB = 4302.57
        wV = 5428.55
        wr = (wavelengths - wB) / (wV - wB)

        numerator = 1.0 * wr
        denominator = 1.0

        wi = wr * wr
        for p in params:
            numerator += wi * p
            denominator += p
            wi *= wr

        return -numerator / denominator



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst


import sys
import imp
try:
    # This incantation forces distribute to be used (over setuptools) if it is
    # available on the path; otherwise distribute will be downloaded.
    import pkg_resources
    distribute = pkg_resources.get_distribution('distribute')
    if pkg_resources.get_distribution('setuptools') != distribute:
        sys.path.insert(1, distribute.location)
        distribute.activate()
        imp.reload(pkg_resources)
except:  # There are several types of exceptions that can occur here
    from distribute_setup import use_setuptools
    use_setuptools()

import glob
import os
from setuptools import setup, find_packages


# Need a recursive glob to find all package data files if there are
# subdirectories
import fnmatch
def recursive_glob(basedir, pattern):
    matches = []
    for root, dirnames, filenames in os.walk(basedir):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches

#A dirty hack to get around some early import/configurations ambiguities
#This is the same as setup_helpers.set_build_mode(), but does not require
#importing setup_helpers
if sys.version_info[0] >= 3:
    import builtins
else:
    import __builtin__ as builtins
builtins._PACKAGE_SETUP_ = True

import astropy
from astropy.setup_helpers import (register_commands, adjust_compiler,
                                   filter_packages, update_package_files,
                                   get_debug_option)
from astropy.version_helpers import get_git_devstr, generate_version_py

# Set affiliated package-specific settings
PACKAGENAME = 'sncosmo'
DESCRIPTION = 'Package for supernova cosmology based on astropy'
LONG_DESCRIPTION = ''
AUTHOR = 'The SNCosmo Developers'
AUTHOR_EMAIL = 'kylebarbary@gmail.com'
LICENSE = 'BSD'
URL = 'http://sncosmo.readthedocs.org'

#VERSION should be PEP386 compatible (http://www.python.org/dev/peps/pep-0386)
VERSION = '0.1'

# Indicates if this version is a release version
RELEASE = 'dev' not in VERSION

if not RELEASE:
    VERSION += get_git_devstr(False)

# Populate the dict of setup command overrides; this should be done before
# invoking any other functionality from distutils since it can potentially
# modify distutils' behavior.
cmdclassd = register_commands(PACKAGENAME, VERSION, RELEASE)

# Adjust the compiler in case the default on this platform is to use a
# broken one.
adjust_compiler(PACKAGENAME)

# Freeze build information in version.py
generate_version_py(PACKAGENAME, VERSION, RELEASE, get_debug_option())

# Use the find_packages tool to locate all packages and modules
packagenames = filter_packages(find_packages())

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if (os.path.basename(fname) != 'README.rst' and fname[-1] != '~')]

# This dictionary stores the command classes used in setup below
#cmdclassd = {'test': setup_helpers.setup_test_command(PACKAGENAME),

             # Use distutils' sdist because it respects package_data.
             # setuptools/distributes sdist requires duplication of
             # information in MANIFEST.in
#             'sdist': sdist.sdist,

             # Use a custom build command which understands additional
             # commandline arguments
#             'build': setup_helpers.AstropyBuild,

             # Use a custom install command which understands additional
             # commandline arguments
#             'install': setup_helpers.AstropyInstall

#             }

#if setup_helpers.HAVE_CYTHON and not RELEASE:
#    from Cython.Distutils import build_ext
    # Builds Cython->C if in dev mode and Cython is present
#    cmdclassd['build_ext'] = setup_helpers.wrap_build_ext(build_ext)
#else:
#    cmdclassd['build_ext'] = setup_helpers.wrap_build_ext()

#if setup_helpers.AstropyBuildSphinx is not None:
#    cmdclassd['build_sphinx'] = setup_helpers.AstropyBuildSphinx

# Set our custom command class mapping in setup_helpers, so that
# setup_helpers.get_distutils_option will use the custom classes.
#setup_helpers.cmdclassd = cmdclassd

# Additional C extensions that are not Cython-based should be added here.
extensions = []

# A dictionary to keep track of all package data to install
data_files = recursive_glob(os.path.join(PACKAGENAME, 'data'), '*')
data_files = [f[len(PACKAGENAME)+1:] for f in data_files]
package_data = {PACKAGENAME: data_files}

# A dictionary to keep track of extra packagedir mappings
package_dirs = {}

# Update extensions, package_data, packagenames and package_dirs from
# any sub-packages that define their own extension modules and package
# data.  See the docstring for setup_helpers.update_package_files for
# more details.
update_package_files(PACKAGENAME, extensions, package_data,
                     packagenames, package_dirs)

setup(name=PACKAGENAME,
      version=VERSION,
      description=DESCRIPTION,
      packages=packagenames,
      package_data=package_data,
      package_dir=package_dirs,
      ext_modules=extensions,
      scripts=scripts,
      requires=['numpy', 'scipy', 'astropy'],
      install_requires=['numpy', 'scipy', 'astropy'],
      provides=[PACKAGENAME],
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url=URL,
      long_description=LONG_DESCRIPTION,
      classifiers=[
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Topic :: Scientific/Engineering :: Physics'
      ],
      cmdclass=cmdclassd,
      zip_safe=False,
      use_2to3=True
      )



import numpy as np
import sncosmo

meta, data = sncosmo.readlc('simulated_salt2_data.dat')

model = sncosmo.get_model('salt2')
model.set(**meta)

sncosmo.plotlc(data, fname='plotlc_example.png', model=model)



import numpy as np
import sncosmo
from astropy.utils import OrderedDict as odict

model = sncosmo.get_model('salt2')
params = odict(z=0.5, c=0.2, t0=55100., mabs=-19.5, x1=0.5)
model.set(**params)

times = np.linspace(55070., 55150., 40)
bands = np.array(10 * ['sdssg', 'sdssr', 'sdssi', 'sdssz'])
zp = 25. * np.ones(40)
zpsys = np.array(40 * ['ab'])

flux = model.bandflux(bands, times, zp=zp, zpsys=zpsys)
fluxerr = (0.05 * np.max(flux)) * np.ones(40, dtype=np.float)
flux += fluxerr * np.random.randn(40)

data = odict([('time', times), ('band', bands), ('flux', flux), 
              ('fluxerr', fluxerr), ('zp', zp), ('zpsys', zpsys)])

sncosmo.writelc(data, 'simulated_salt2_data.dat', meta=params)
